# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import logging
import os
import re
from io import BytesIO
from typing import Any, Dict, List, Optional, Union

import lxml.etree as ET
import numpy as np
import pandas as pd

from opcua_tools import ua_models
from opcua_tools.ua_data_types import UANodeId
from opcua_tools.value_parser import parse_nodeid, parse_value

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

tagsplit = re.compile(r"({.*\})(.*)")
OPCFOUNDATION_NAMESPACE = "http://opcfoundation.org/UA/"


def findrefs(
    elem, uaxsd, namespace_map: Dict[int, int], alias_map: Dict[str, UANodeId]
):
    return [
        (
            parse_nodeid(r.text.rstrip(), namespace_map, alias_map),
            fix_ref_attrib(r, namespace_map, alias_map),
        )
        for refs in elem.findall(uaxsd + "References")
        for r in refs.findall(uaxsd + "Reference")
    ]


def findval(elem, uaxsd):
    val = elem.find(uaxsd + "Value")
    if val is None:
        return pd.NA
    if len(val) < 1:
        return pd.NA

    return parse_value(val)


def finddisplayname(elem, uaxsd):
    val = elem.find(uaxsd + "DisplayName")
    if val is None:
        return ""

    if val.text is None:
        return ""

    return val.text.rstrip()


def finddescription(elem, uaxsd):
    val = elem.find(uaxsd + "Description")
    if val is None:
        return ""

    if val.text is None:
        return ""

    return val.text.rstrip()


def parse_node_attrib(
    elem: ET.ElementBase, namespace_map: Dict[int, int], alias_map: Dict[str, UANodeId]
) -> Dict[str, str]:
    """
    Parses the xml element for the node and maps NodeId, DataType (if any), ParentNodeID(if any)
    in the Attribute Dictionary.

    Parameters:
    elem (XML element): the XML element to parse
    namespace_map (Dict[str,str]): The dictionary used for mapping of ns index

    Returns:
    Dict[str,str] with attributes according to NodeClass in xml element
    """
    attrib = dict(elem.attrib)
    # Map NodeId
    attrib["NodeId"] = parse_nodeid(attrib["NodeId"], namespace_map, alias_map)
    if "DataType" in attrib:
        attrib["DataType"] = parse_nodeid(attrib["DataType"], namespace_map, alias_map)
    if "ParentNodeId" in attrib:
        attrib["ParentNodeId"] = parse_nodeid(
            attrib["ParentNodeId"], namespace_map, alias_map
        )
    if "MethodDeclarationId" in attrib:
        attrib["MethodDeclarationId"] = parse_nodeid(
            attrib["MethodDeclarationId"], namespace_map, alias_map
        )

    return attrib


def fix_ref_attrib(
    elem: ET.ElementBase, namespace_map: Dict[int, int], alias_map: Dict[str, UANodeId]
) -> Dict[str, str]:
    """
    Parses the xml element for the reference and maps ReferenceType  in the attribute Dictionary.

    Parameters:
    elem (XML element): the XML element to parse
    namespace_map (Dict[str,str]): The dictionary used for mapping of ns index

    Returns:
    Dict[str,str] with attributes according to the reference xml element
    """
    attrib = dict(elem.attrib)
    attrib["ReferenceType"] = parse_nodeid(
        attrib["ReferenceType"], namespace_map, alias_map=alias_map
    )
    return attrib


def process_elem_batch(
    elems: List[ET.ElementBase],
    uaxsd: str,
    namespace_map: Dict[int, int],
    alias_map: Dict[str, UANodeId],
) -> pd.DataFrame:
    """
    Creates a Pandas DataFrame with node and reference info based on the provided XML element list

    Parameters:

    Returns:
    Pandas DataFrame with columns:[Tag,Attrib,DisplayName, Description, References, ValueTmp]

    """
    df = pd.DataFrame({"elem": elems})
    df["Tag"] = df["elem"].map(lambda x: x.tag).str.replace(uaxsd, "", regex=True)
    df["Attrib"] = df["elem"].map(
        lambda x: parse_node_attrib(x, namespace_map, alias_map)
    )
    df["DisplayName"] = df["elem"].map(lambda x: finddisplayname(x, uaxsd))
    df["Description"] = df["elem"].map(lambda x: finddescription(x, uaxsd))
    df["References"] = df["elem"].map(
        lambda x: findrefs(x, uaxsd, namespace_map, alias_map)
    )
    df["Value"] = df["elem"].map(lambda x: findval(x, uaxsd))
    df = df.drop(columns="elem")
    df = df.convert_dtypes()
    return df


def iterparse_xml(
    xmlfile: str, desired_namespace_list: List[str], batchsize=100000
) -> Dict[str, Any]:
    """
    Parses the XML file to a dictionary with a nodes pd.DataFrame and an aliaslist with refrencetype aliasnames

    Parameters:
    xmlfile (str): Path to file to read.
    desired_namespace_list (List[str]): The list with namspace uris in the desired order.

    Returns:
    Dictionary (str,object): The with dictionary with pd.Dataframe in 'nodes' key and alias dict in 'alias_map' key
    Node Dataframe example:
     Tag                                             Attrib                                          References                                           ValueTmp
    ['UAVariable'    {'NodeId': '0:0:8244', 'BrowseName': 'Annotation', 'ParentNodeId': '0:0:7617', 'DataType': 'String'}   'Annotation' None

    list([('0:0:69', {'ReferenceType': 'HasTypeDefinition'}), ('0:0:7617', {'ReferenceType': 'HasComponent', 'IsForward': 'false'})])

    """
    uaxsd = "{http://opcfoundation.org/UA/2011/03/UANodeSet.xsd}"

    tags_to_find = list(
        map(
            lambda x: uaxsd + x,
            ["NamespaceUris", "Uri", "Model", "RequiredModel", "Alias"],
        )
    )
    namespace_list = []
    alias_map = {}

    nodeclasses = [
        "UAObjectType",
        "UAObject",
        "UAVariableType",
        "UAVariable",
        "UADataType",
        "UAReferenceType",
        "UAView",
        "UAMethod",
    ]
    nodeclasses_xsd = list(map(lambda x: uaxsd + x, nodeclasses))

    nodeset = uaxsd + "UANodeSet"

    tagiter = ET.iterparse(
        xmlfile,
        events=("start", "end"),
        tag=[nodeset] + nodeclasses_xsd + tags_to_find,
        encoding="utf-8",
    )

    elems = []
    df_list = []
    i = 1
    foundnses = False
    namespace_map = {0: 0}
    models = []
    current_model = None

    for event, elem in tagiter:
        if elem.tag == nodeset:
            if event == "start":
                elem.nsmap.copy()
            if event == "end":
                pass
        elif not foundnses and event == "end" and elem.tag == uaxsd + "Uri":
            namespace_list.append(elem.text)
            elem.clear()
        elif not foundnses and event == "end" and elem.tag == uaxsd + "NamespaceUris":
            # All uris in "NamespaceUris" are parsed
            foundnses = True
            extend_namespace_map(desired_namespace_list, namespace_list, namespace_map)
        elif elem.tag == f"{uaxsd}Model":
            if event == "start":
                model = ua_models.UAModel(
                    model_uri=elem.attrib["ModelUri"],
                    version=elem.attrib.get("Version"),
                    publication_date=elem.attrib.get("PublicationDate"),
                )
                current_model = model
                models.append(model)
        elif elem.tag == f"{uaxsd}RequiredModel":
            if event == "start":
                required_model = ua_models.UARequiredModel(
                    model_uri=elem.attrib["ModelUri"],
                    version=elem.attrib.get("Version"),
                    publication_date=elem.attrib.get("PublicationDate"),
                )
                current_model.required_models.append(required_model)
        elif event == "end" and elem.tag == uaxsd + "Alias":
            alias_map[elem.attrib["Alias"]] = parse_nodeid(elem.text, namespace_map)
            elem.clear()
        elif event == "end":
            elems.append(elem)
        if i % batchsize == 0:
            logger.info(f"Processing XML node batch {i}")
            df = process_elem_batch(
                elems=elems,
                uaxsd=uaxsd,
                namespace_map=namespace_map,
                alias_map=alias_map,
            )
            df_list.append(df)
            # Release memory
            list(map(lambda x: x.clear(), elems))
            elems = []
        i = i + 1

    if len(elems) > 0:
        df = process_elem_batch(elems, uaxsd, namespace_map, alias_map)
        # Release memory
        list(map(lambda x: x.clear(), elems))
        df_list.append(df)

    nodes = pd.concat(df_list)

    return {
        "nodes": nodes,
        "alias_map": alias_map,
        "namespace_map": namespace_map,
        "models": models,
    }


def extend_namespace_map(
    existing_namespaces: List[str],
    namespace_list: List[str],
    namespace_map: Dict[int, int],
) -> None:
    """
    Extends a dictionary with ns mapping based on the order from desired_namespace_list.

    Parameters:
    desired_namespace_list (List[str]) : The list of namespaces to map to
    namespace_list (List[str]) : The list of namespaces to map from

    Returns:
    Dictionary (str,str): The mapping dictionary based on Namespace index: "From Idx":"To Idx"
    """
    if 0 not in namespace_map:
        namespace_map[0] = 0
    for i, n in enumerate(namespace_list):
        if n not in existing_namespaces:
            logger.warning(
                "Namespace "
                + n
                + " not found in namespace list, adding with index "
                + str(len(existing_namespaces))
            )
            existing_namespaces.append(n)

        namespace_map[i + 1] = existing_namespaces.index(n)


def normalize_wrt_nodeid(nodes: pd.DataFrame, references: pd.DataFrame) -> pd.DataFrame:
    logger.info("Started normalizing table structure with respect to nodeid")

    nodecols = ["NodeId", "ParentNodeId", "DataType", "MethodDeclarationId"]
    refcols = ["Src", "Trg", "ReferenceType"]

    allids = pd.concat(
        [nodes[c].dropna() for c in nodecols if c in nodes.columns.values]
        + [references[c].dropna() for c in refcols if c in references.columns.values],
        ignore_index=True,
    )
    codes, uniques = pd.factorize(allids)

    lookup_df = pd.DataFrame({"uniques": uniques})

    uniques_index = pd.Index(uniques)
    convert_to_int_index = lambda x: uniques_index.get_indexer(pd.Index(x)).astype(
        pd.Int32Dtype
    )
    for c in refcols:
        references[c] = convert_to_int_index(references[c])

    nodes["id"] = convert_to_int_index(nodes["NodeId"])
    for c in nodecols[1:]:  # Skip nodeids!
        if c in nodes.columns.values:
            nodes[c] = convert_to_int_index(nodes[c])
            nodes[c] = nodes[c].replace(-1, pd.NA)

    logger.info("Finished normalizing table structure with respect to nodeid")
    return lookup_df


def get_xml_namespaces(xml_file: str) -> List[str]:
    """Opens the xml file provided and peeks inside the file to look for any tags which
    indicate the namespace uri of the xml file.

    Args:
        xml_file (str): The full filepath to the xml file to check

    Returns:
        List[str] containing the different namespace uris which are found in the file.

    """
    uaxsd = "{http://opcfoundation.org/UA/2011/03/UANodeSet.xsd}"

    namespace_list = []
    # In some NodeSet2 definition files the <Model> tag is not found
    # therefore using the common files name as well.
    if xml_file.endswith("Opc.Ua.NodeSet2.xml"):
        namespace_list.append("http://opcfoundation.org/UA")
        namespace_list.append(OPCFOUNDATION_NAMESPACE)
        return namespace_list

    # Adding tags which contain Models and ModelUri.
    tree = ET.parse(xml_file)
    root = tree.getroot()

    found_nses = False
    root_iter_models = root.iter(uaxsd + "Models")
    if root_iter_models:
        for models_tag in root_iter_models:
            for model in models_tag.iter(uaxsd + "Model"):
                model_uri = model.get("ModelUri")
                if not found_nses and model_uri:
                    namespace_list.append(model_uri)
        return namespace_list

    # Adding tags which contain NamespaceUris
    tag_namespace = ET.iterparse(
        xml_file, events=("start", "end"), tag=[uaxsd + "Uri", uaxsd + "NamespaceUris"]
    )

    found_nses = False
    for event, elem in tag_namespace:
        if not found_nses and event == "end" and elem.tag == uaxsd + "Uri":
            namespace_list.append(elem.text)
            elem.clear()
        elif not found_nses and event == "end" and elem.tag == uaxsd + "NamespaceUris":
            # All uris in "NamespaceUris" are parsed
            break

    return namespace_list


def get_namespace_data_from_file(xml_file: str) -> dict:
    """Opens the xml file provided and peeks inside the file to look for Namespace tags
    to resolve which namespaces it includes.

    Args:
        xml_file (str): The full filepath to the xml file to check

    Returns:
        A dictionary containing file namespace name and a list of required namespaces
        Example:
            {
                "name": "http://powerview.com/enterprise",
                "included_namespaces": {
                    "http://prediktor.no/PVTypes/",
                    "http://opcfoundation.org/UA/",
                }
            }
    """
    uaxsd = "{http://opcfoundation.org/UA/2011/03/UANodeSet.xsd}"

    # In some NodeSet2 definition files the <Model> tag is not found
    # therefore using the common files name as well.
    opcua_namespace_data = {
        "name": OPCFOUNDATION_NAMESPACE,
        "included_namespaces": set(),
    }
    if xml_file.endswith("Opc.Ua.NodeSet2.xml"):
        return opcua_namespace_data

    tree = ET.parse(xml_file)
    root = tree.getroot()

    root_iter_models = root.iter(uaxsd + "Models")

    namespace_data = None
    for models_tag in root_iter_models:
        for model in models_tag:
            model_uri = model.get("ModelUri")
            if model_uri == OPCFOUNDATION_NAMESPACE:
                namespace_data = opcua_namespace_data
            else:
                namespace_data = {
                    "name": model.get("ModelUri"),
                    "included_namespaces": {OPCFOUNDATION_NAMESPACE},
                }
            model.clear()
            break  # get just the first Model tag (there should be no more Model tags)
        models_tag.clear()

    if namespace_data is None:
        message = f"Missing 'Model' tag in {xml_file}"
        logger.error(message)
        raise ValueError(message)

    root_namespace_uris = root.iter(uaxsd + "NamespaceUris")
    for namespace_uris_tag in root_namespace_uris:
        for uri_tag in namespace_uris_tag:
            namespace_uri = uri_tag.text
            if namespace_uri != namespace_data["name"]:
                namespace_data["included_namespaces"].add(namespace_uri)
            uri_tag.clear()
        namespace_uris_tag.clear()

    return namespace_data


def get_list_of_xml_files(xml_directory_path: str) -> List[str]:
    """Returns a list of xml_files in a provided directory

    Args:
        xml_directory_path (str): Full path to the directory

    Return:
        List[str] of xmls which are found in the directory

    """

    files = []
    for file in os.listdir(xml_directory_path):
        full_path = os.path.join(xml_directory_path, file)
        if os.path.isfile(full_path) and full_path.endswith(".xml"):
            files.append(full_path)

    return files


def exclude_files_not_in_namespaces(
    input_files: List[str], namespaces: List[str]
) -> List[str]:
    """Removes the files which do not have any NamespaceUris in the namespaces list.

    Args:
        input_files (List[str]): List of files to filter through.
        namespaces (List[str]): List of desired namespaces to keep.

    Return:
        List[str] of xmls which are found in the directory

    """
    # Removes None for the list if found in the namespaces
    namespaces_set_list = list(set(filter(None, namespaces)))
    output_files = input_files.copy()
    for file in input_files:
        file_ns_uris = get_xml_namespaces(file)
        # Removes file if none of the file_ns_uris are in namespaces_set_list
        if not any(file in file_ns_uris for file in namespaces_set_list):
            output_files.remove(file)

    return output_files


def parse_xml(
    xmlfile: Union[str, BytesIO], namespaces: Optional[List[str]] = None
) -> Dict[str, Any]:
    parse_dict = parse_xml_without_normalization(xmlfile, namespaces)
    lookup_df = normalize_wrt_nodeid(parse_dict["nodes"], parse_dict["references"])
    parse_dict["lookup_df"] = lookup_df
    return parse_dict


def parse_xml_without_normalization(
    xmlfile: Union[str, BytesIO], namespaces: Optional[List[str]] = None
) -> Dict[str, Any]:
    if namespaces is None:
        namespaces = []

    uans = OPCFOUNDATION_NAMESPACE
    if uans not in namespaces:
        namespaces.append(uans)

    parse_dict = iterparse_xml(xmlfile, namespaces)
    nodes = parse_dict["nodes"].reset_index()
    nodes = nodes.rename(columns={"Tag": "NodeClass"})

    attrib_df = get_attrib_df(nodes)
    attrib_df = attrib_df.fillna(pd.NA)
    nodes = pd.concat([nodes, attrib_df], axis=1).drop(columns="Attrib")

    references = nodes.loc[:, ["NodeId", "References"]]
    references = (
        references.rename(columns={"NodeId": "Src"})
        .explode("References")
        .reset_index(drop=True)
    )
    references = references.loc[~references["References"].isna()]

    references["Trg"] = references["References"].map(lambda x: x[0])
    references["IsForward"] = references["References"].map(
        lambda x: (
            False if "IsForward" in x[1] and x[1]["IsForward"] == "false" else True
        )
    )
    refsrc = references[["Src"]]
    references.loc[~references["IsForward"], "Src"] = references.loc[
        ~references["IsForward"], "Trg"
    ]
    references.loc[~references["IsForward"], "Trg"] = refsrc.loc[
        ~references["IsForward"], "Src"
    ]

    references["ReferenceType"] = references["References"].map(
        lambda x: x[1]["ReferenceType"]
    )
    references = references.drop(columns=["References", "IsForward"])
    references = references.drop_duplicates().reset_index(drop=True)
    nodes = nodes.drop(columns=["References", "index"])

    # Browsenames are prefixed with namespace indices, but this makes for difficult to follow browsepaths.
    # We keep Browsename namespaces in its own column in order not to lose this information
    namespace_map = parse_dict["namespace_map"]
    nodes["BrowseNameNamespace"] = nodes["BrowseName"].map(
        lambda x: int(x.split(":")[0]) if ":" in x else 0
    )
    nodes["BrowseNameNamespace"] = nodes["BrowseNameNamespace"].map(namespace_map)
    nodes["BrowseName"] = (
        nodes["BrowseName"]
        .map(lambda x: x.split(":")[1] if ":" in x else x)
        .astype("str")
    )

    if "Value" not in nodes.columns.values:
        nodes["Value"] = pd.NA

    nodes["ns"] = nodes["NodeId"].map(lambda x: x.namespace).astype(pd.Int8Dtype())
    models = parse_dict["models"]
    return {
        "nodes": nodes,
        "references": references,
        "namespaces": namespaces,
        "models": models,
    }


def get_attrib_df(nodes: pd.DataFrame) -> pd.DataFrame:
    attrib_df = pd.DataFrame.from_records(nodes["Attrib"].values)

    is_abstract_column_name = "IsAbstract"
    if is_abstract_column_name in attrib_df.columns:
        attrib_df[is_abstract_column_name] = attrib_df[is_abstract_column_name].replace(
            {np.nan: False, "false": False}
        )
        attrib_df[is_abstract_column_name] = attrib_df[is_abstract_column_name].astype(
            "bool"
        )

    symmetric_column_name = "Symmetric"
    if symmetric_column_name in attrib_df.columns:
        attrib_df[symmetric_column_name] = attrib_df[symmetric_column_name].replace(
            {np.nan: False, "false": False}
        )
        attrib_df[symmetric_column_name] = attrib_df[symmetric_column_name].astype(
            "bool"
        )

    attrib_df = attrib_df.convert_dtypes()

    value_rank_column_name = "ValueRank"
    if value_rank_column_name in attrib_df.columns:
        attrib_df[value_rank_column_name] = attrib_df[value_rank_column_name].astype(
            "Int8"
        )

    minimum_sampling_interval_column_name = "MinimumSamplingInterval"
    if minimum_sampling_interval_column_name in attrib_df.columns:
        attrib_df[minimum_sampling_interval_column_name] = attrib_df[
            minimum_sampling_interval_column_name
        ].astype("Int32")

    access_level_column_name = "AccessLevel"
    if access_level_column_name in attrib_df.columns:
        attrib_df[access_level_column_name] = attrib_df[
            access_level_column_name
        ].astype("Int8")

    event_notifier_column_name = "EventNotifier"
    if event_notifier_column_name in attrib_df.columns:
        attrib_df["EventNotifier"] = attrib_df["EventNotifier"].astype("Int8")

    return attrib_df


def parse_xml_dir(
    xml_dir: str, namespaces: Optional[List[str]] = None
) -> Dict[str, Any]:
    """Parses xml directory and creates Pandas tables which contains
    a consolidated nodes, references, namespaces, and lookup_df
    for the files found in the nodes. It will create a list of xml files
    contained in the xml directory. If the 'namespaces' parameter is
    provided, it will not include files with namespaces which are not
    found in the namespace list.

    Args:
        xml_dir (str): Full path to the directory of xmls
        namespaces (Optional[List[str]] = None): Dictionary of namespaces

    Return:
        Dict[str, Any] of parsed pandas tables from the xml directories.

    """

    files = get_list_of_xml_files(xml_dir)
    return parse_xml_files(files, namespaces)


def parse_xml_files(
    files: List[str], namespaces: Optional[List[str]] = None
) -> Dict[str, Any]:
    """Parses xml list of files and creates Pandas tables which contains
    a consolidated nodes, references, namespaces, and lookup_df
    for the files found in the nodes. It will create a list of xml files
    contained in the xml directory. If the 'namespaces' parameter is
    provided, it will not include files with namespaces which are not
    found in the namespace list.

    Args:
        files (str): Full path of xml files
        namespaces (Optional[List[str]] = None): Dictionary of namespaces

    Return:
        Dict[str, Any] of parsed pandas tables from the xml directories.

    """

    if namespaces:
        files = exclude_files_not_in_namespaces(files, namespaces)

    if namespaces is None:
        namespaces = []
    df_nodes_list = []
    df_references_list = []

    files.sort()
    models = []

    for file in files:
        if not file.endswith(".xml"):
            continue
        if not os.path.exists(file):
            raise FileNotFoundError(
                "The specified xml file does not exist or incorrect path was provided"
            )

        logger.info("Started parsing " + str(file))
        parse_dict = parse_xml_without_normalization(file, namespaces)
        namespaces = parse_dict["namespaces"]
        df_nodes_list.append(parse_dict["nodes"])
        df_references_list.append((parse_dict["references"]))
        models = list(itertools.chain(models, parse_dict["models"]))
        logger.info("Finished parsing " + str(file))

    nodes = pd.concat(df_nodes_list, ignore_index=True)
    columns_to_fix_missing_values = [
        "SymbolicName",
        "IsAbstract",
        "Symmetric",
        "ValueRank",
        "ParentNodeId",
        "ArrayDimensions",
        "ReleaseStatus",
        "MinimumSamplingInterval",
        "MethodDeclarationId",
        "EventNotifier",
    ]
    for column in columns_to_fix_missing_values:
        if column in nodes.columns:
            nodes[column] = nodes[column].replace({np.nan: pd.NA})

    references = pd.concat(df_references_list, ignore_index=True)
    lookup_df = normalize_wrt_nodeid(nodes, references)

    return {
        "nodes": nodes,
        "references": references,
        "namespaces": namespaces,
        "lookup_df": lookup_df,
        "models": models,
    }



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from pathlib import Path
from unittest import mock

import pandas as pd
from definitions import get_project_root

import opcua_tools as ot
from opcua_tools import UAGraph, ua_models
from opcua_tools.nodeset_generator import (
    denormalize_nodes_nodeids,
    denormalize_references_nodeids,
)

PATH_HERE = os.path.dirname(__file__)


def test_idempotency():
    parse_dict = ot.parse_xml_dir(PATH_HERE + "/testdata/generator")
    ot.create_nodeset2_file(
        nodes=parse_dict["nodes"].copy(),
        references=parse_dict["references"].copy(),
        models=parse_dict["models"].copy(),
        namespaces=parse_dict["namespaces"],
        serialize_namespace=0,
        filename_or_stringio=PATH_HERE + "/expected/generator/nodeset2.xml",
    )
    parse_dict2 = ot.parse_xml_dir(PATH_HERE + "/expected/generator")

    nodes = parse_dict["nodes"]
    references = parse_dict["references"]
    nodes = denormalize_nodes_nodeids(nodes, parse_dict["lookup_df"])
    references = denormalize_references_nodeids(references, parse_dict["lookup_df"])
    nodes = nodes.drop(columns=["id"])
    nodes = nodes.sort_values(by=nodes.columns.values.tolist()).reset_index(drop=True)
    references = references.sort_values(
        by=references.columns.values.tolist()
    ).reset_index(drop=True)

    nodes2 = parse_dict2["nodes"]
    references2 = parse_dict2["references"]
    nodes2 = denormalize_nodes_nodeids(nodes2, parse_dict2["lookup_df"])
    references2 = denormalize_references_nodeids(references2, parse_dict2["lookup_df"])
    nodes2 = nodes2.drop(columns=["id"])
    nodes2 = nodes2.sort_values(by=nodes.columns.values.tolist()).reset_index(drop=True)
    references2 = references2.sort_values(
        by=references.columns.values.tolist()
    ).reset_index(drop=True)

    # print(set(nodes['Value'].tolist()).difference(set(nodes2['Value'].tolist())))

    # print(nodes[nodes['DataType'].values != nodes2['DataType'].values][['NodeId', 'DataType']])
    # print(nodes2[nodes2['DataType'].values != nodes['DataType'].values][['NodeId', 'DataType']])

    assert set(nodes.columns.values.tolist()) == set(nodes2.columns.values.tolist())

    nodes2 = nodes2[nodes.columns.values.tolist()].copy()
    # nodes2.to_csv('nodes2.csv')
    # nodes.to_csv('nodes.csv')
    pd.testing.assert_frame_equal(nodes, nodes2)
    pd.testing.assert_frame_equal(references, references2)

    os.remove(PATH_HERE + "/expected/generator/nodeset2.xml")


def test_ua_graph_write_nodeset_without_crash(paper_example_path):
    path_to_xmls = str(paper_example_path)
    ua_graph = UAGraph.from_path(path_to_xmls)

    output_folder = get_project_root() / "tests" / "output"
    output_file_path = str(output_folder / "paper_example_output.xml")

    # Creating output folder if it does not exist
    if not os.path.exists(str(output_folder)):
        os.makedirs(str(output_folder))

    ua_graph.write_nodeset(output_file_path, "http://prediktor.com/paper_example")


@mock.patch("opcua_tools.nodeset_generator.create_required_models")
def test_ua_graph_write_nodeset_with_required_models(create_required_models_mock):
    path_to_xml = str(
        Path(PATH_HERE) / "testdata" / "parser" / "Opc.Ua.IEC61850-6.NodeSet2.xml"
    )
    path_to_type_library = str(
        Path(PATH_HERE) / "testdata" / "parser" / "Opc.Ua.NodeSet2.xml"
    )
    ua_graph = UAGraph.from_file_list([path_to_xml, path_to_type_library])

    output_folder = get_project_root() / "tests" / "output"
    output_file_path = str(output_folder / "required_models_output.xml")

    # Creating output folder if it does not exist
    if not os.path.exists(str(output_folder)):
        os.makedirs(str(output_folder))

    namespace_uri = "http://opcfoundation.org/UA/IEC61850-6"
    ua_graph.write_nodeset(output_file_path, namespace_uri)

    create_required_models_mock.assert_called_once_with(
        ua_models.UAModel(
            model_uri=namespace_uri,
            publication_date="2018-02-05T00:00:00Z",
            version="2.0",
            required_models=[
                ua_models.UARequiredModel(
                    model_uri="http://opcfoundation.org/UA/IEC61850-7-3",
                    publication_date=None,
                    version="2.0",
                ),
                ua_models.UARequiredModel(
                    model_uri="http://opcfoundation.org/UA/",
                    publication_date="2019-05-01T00:00:00Z",
                    version="1.04",
                ),
            ],
        )
    )



import os

import pandas as pd
import pytest
from definitions import get_project_root

from opcua_tools.ua_data_types import UABoolean, UAInt32
from opcua_tools.ua_graph import UAGraph
from opcua_tools.validator import exceptions, value_validator


class TestValidatorFeature:
    def test_validator_raises_error_when_data_type_is_missing_for_uavariable(
        self, invalid_files_root_path
    ):
        path_to_xmls = str(invalid_files_root_path / "missing_data_types")
        ua_graph = UAGraph.from_path(path_to_xmls)

        output_folder = get_project_root() / "tests" / "output"
        output_file_path = str(
            output_folder / "invalid_output_that_eventually_wont_be_created.xml"
        )

        # Creating output folder if it does not exist
        if not os.path.exists(str(output_folder)):
            os.makedirs(str(output_folder))

        with pytest.raises(exceptions.ValidationError):
            ua_graph.write_nodeset(
                output_file_path, "http://prediktor.com/paper_example"
            )

    def test_validate_values_in_df_raises_validation_error(
        self, invalid_files_root_path
    ):
        path_to_xmls = str(invalid_files_root_path / "non_matching_datatype")
        ua_graph = UAGraph.from_path(path_to_xmls)

        output_folder = get_project_root() / "tests" / "output"
        output_file_path = str(
            output_folder / "invalid_output_that_eventually_wont_be_created.xml"
        )

        # Creating output folder if it does not exist
        if not os.path.exists(str(output_folder)):
            os.makedirs(str(output_folder))

        with pytest.raises(exceptions.ValidationError) as e:
            ua_graph.write_nodeset(
                output_file_path, "http://prediktor.com/paper_example"
            )
        exception_obj = e.value
        assert (
            str(exception_obj)
            == "Invalid Value for rows with the following display names: ['FunctionalAspectName']."
        )


class TestValidatorUnit:
    def test_validator_df_is_valid_when_empty(self, sample_date_type_nodes):
        empty_df = pd.DataFrame({})

        result_df = value_validator.validate_values_in_df(
            empty_df, sample_date_type_nodes
        )

        pd.testing.assert_frame_equal(result_df, empty_df)

    def test_validator_df_is_valid_when_nodeclass_is_not_uavariable(
        self, ua_object_node_df, sample_date_type_nodes
    ):
        ua_object_node_df["NodeClass"] = "UAObject"

        result_df = value_validator.validate_values_in_df(
            ua_object_node_df, sample_date_type_nodes
        )

        pd.testing.assert_frame_equal(result_df, ua_object_node_df)

    def test_validator_df_is_valid_when_uavariable_has_empty_value(
        self, ua_object_node_df, sample_date_type_nodes
    ):
        result_df = value_validator.validate_values_in_df(
            ua_object_node_df, sample_date_type_nodes
        )

        pd.testing.assert_frame_equal(result_df, ua_object_node_df)

    def test_validator_row_is_invalid_when_opcua_tools_class_name_is_invalid(
        self, ua_object_node_df, sample_date_type_nodes
    ):
        ua_object_node_df["Value"] = UABoolean(value=True)
        invalid_data_type_id = 1
        ua_object_node_df["DataType"] = invalid_data_type_id

        with pytest.raises(exceptions.ValidationError):
            value_validator.validate_values_in_df(
                ua_object_node_df, sample_date_type_nodes
            )

    def test_validator_row_is_valid_when_enumeration_is_given(
        self, ua_object_node_df, sample_date_type_nodes
    ):
        ua_object_node_df["Value"] = UAInt32()
        ua_object_node_df["DataType"] = 0

        result_df = value_validator.validate_values_in_df(
            ua_object_node_df, sample_date_type_nodes
        )

        pd.testing.assert_frame_equal(result_df, ua_object_node_df)

    def test_validator_gets_correct_data_type(
        self, ua_object_node_df, sample_date_type_nodes
    ):
        ua_object_node_df["Value"] = UAInt32()
        sample_nodes_int32_id = 1
        ua_object_node_df["DataType"] = sample_nodes_int32_id

        result_df = value_validator.validate_values_in_df(
            ua_object_node_df, sample_date_type_nodes
        )

        pd.testing.assert_frame_equal(result_df, ua_object_node_df)



"""
Call the memory optimizer in order to minimize possible pandas' memory leaks.
Works for Linux. Does not change anything for MacOS.
"""

import logging
from ctypes import CDLL, cdll

import pandas as pd

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


__old_pandas_del = getattr(pd.DataFrame, "__del__", None)

try:
    cdll.LoadLibrary("libc.so.6")
    libc = CDLL("libc.so.6")
    libc.malloc_trim(0)
except (OSError, AttributeError):
    libc = None


def __malloc_trim_del(self):
    if __old_pandas_del:
        __old_pandas_del(self)
    libc.malloc_trim(0)


def replace_default_pandas_del_method():
    if libc:
        logger.info("Applying memory optimizer for pd.DataFrame.__del__")
        pd.DataFrame.__del__ = __malloc_trim_del
    else:
        logger.info("Memory optimizer not applied: libc or malloc_trim() not found")



import dataclasses
from typing import List, Union


@dataclasses.dataclass
class UAModelBase:
    model_uri: str
    publication_date: Union[str, None]
    version: Union[str, None]


@dataclasses.dataclass
class UARequiredModel(UAModelBase):
    def __hash__(self):
        key_to_hash = "{}_{}_{}".format(
            self.model_uri, self.publication_date, self.version
        )
        return hash(key_to_hash)


@dataclasses.dataclass
class UAModel(UAModelBase):
    required_models: List[UARequiredModel] = dataclasses.field(default_factory=list)



from opcua_tools.ua_data_types import (
    UABoolean,
    UAByte,
    UAByteString,
    UADataValue,
    UADateTime,
    UADiagnosticInfo,
    UADouble,
    UAExpandedNodeId,
    UAExtensionObject,
    UAFloat,
    UAGuid,
    UAInt16,
    UAInt32,
    UAInt64,
    UALocalizedText,
    UANodeId,
    UAQualifiedName,
    UASByte,
    UAStatusCode,
    UAString,
    UAUInt16,
    UAUInt32,
    UAUInt64,
    UAVariant,
    UAXMLElement,
    VariantType,
)

DATA_TYPES_MAPPING = {
    UABoolean.__name__: VariantType.Boolean.name,
    UASByte.__name__: VariantType.SByte.name,
    UAByte.__name__: VariantType.Byte.name,
    UAInt16.__name__: VariantType.Int16.name,
    UAUInt16.__name__: VariantType.UInt16.name,
    UAInt32.__name__: VariantType.Int32.name,
    UAUInt32.__name__: VariantType.UInt32.name,
    UAInt64.__name__: VariantType.Int64.name,
    UAUInt64.__name__: VariantType.UInt64.name,
    UAFloat.__name__: VariantType.Float.name,
    UADouble.__name__: VariantType.Double.name,
    UAString.__name__: VariantType.String.name,
    UADateTime.__name__: VariantType.DateTime.name,
    UAGuid.__name__: VariantType.Guid.name,
    UAByteString.__name__: VariantType.ByteString.name,
    UAXMLElement.__name__: VariantType.XmlElement.name,
    UANodeId.__name__: VariantType.NodeId.name,
    UAExpandedNodeId.__name__: VariantType.ExpandedNodeId.name,
    UAStatusCode.__name__: VariantType.StatusCode.name,
    UAQualifiedName.__name__: VariantType.QualifiedName.name,
    UALocalizedText.__name__: VariantType.LocalizedText.name,
    UAExtensionObject.__name__: VariantType.ExtensionObject.name,
    UADataValue.__name__: VariantType.DataValue.name,
    UAVariant.__name__: VariantType.Variant.name,
    UADiagnosticInfo.__name__: VariantType.DiagnosticInfo.name,
}



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union

import numpy as np
import pandas as pd
from scipy.sparse import coo_matrix, eye

from opcua_tools.value_parser import parse_nodeid


def supertypes_of_nodes(
    types_to_supertype: List[str],
    type_nodes: pd.DataFrame,
    type_references: pd.DataFrame,
) -> pd.DataFrame:
    # Computes the closure
    closure = typing_transitive_reflexive(type_nodes, type_references)
    closure = closure.set_index("Trg")
    typeindex = pd.Index(types_to_supertype)
    closure = closure[closure.index.isin(typeindex)].reset_index()
    closure = closure.rename(columns={"Src": "supertype", "Trg": "type"})

    return closure


def hierarchical_references_trg_has_modelling_rule(
    references: pd.DataFrame, type_references: pd.DataFrame, type_nodes: pd.DataFrame
):
    hierarchical_refs = hierarchical_references(
        inst_references=references,
        type_references=type_references,
        type_nodes=type_nodes,
    )
    has_modelling_rule = has_modelling_rule_references(
        inst_references=references,
        type_references=type_references,
        type_nodes=type_nodes,
    )[["Src"]].rename(columns={"Src": "id"})
    hierarchical_refs_has_modelling_rule = hierarchical_refs.set_index("Trg").join(
        has_modelling_rule.set_index("id"), how="inner"
    )
    hierarchical_refs_has_modelling_rule = (
        hierarchical_refs_has_modelling_rule.reset_index().rename(
            columns={"index": "Trg"}
        )
    )
    return hierarchical_refs_has_modelling_rule[["Src", "Trg", "ReferenceType"]].copy()


# Useful for keeping forward refs for methods when instantiating
def hierarchical_references_trg_has_no_modelling_rule(
    references: pd.DataFrame, type_references: pd.DataFrame, type_nodes: pd.DataFrame
):
    hierarchical_refs = hierarchical_references(
        inst_references=references,
        type_references=type_references,
        type_nodes=type_nodes,
    )
    has_modelling_rule = has_modelling_rule_references(
        inst_references=references,
        type_references=type_references,
        type_nodes=type_nodes,
    )[["Src"]].rename(columns={"Src": "id"})
    has_modelling_rule = has_modelling_rule.set_index("id")
    hierarchical_refs_has_no_modelling_rule = hierarchical_refs.set_index(
        "Trg", drop=False
    )
    trg_no_modelling_rule = ~hierarchical_refs_has_no_modelling_rule.index.isin(
        has_modelling_rule.index
    )
    hierarchical_refs_has_no_modelling_rule = hierarchical_refs_has_no_modelling_rule[
        trg_no_modelling_rule
    ].copy()
    return hierarchical_refs_has_no_modelling_rule[
        ["Src", "Trg", "ReferenceType"]
    ].copy()


def non_hierarchical_references_trg_has_no_modelling_rule(
    references: pd.DataFrame, type_references: pd.DataFrame, type_nodes: pd.DataFrame
) -> pd.DataFrame:
    """This function expects a table of references on which to find the subset
    of references containing only references of the subtype "NonHierarchicalReferences" in
    OPC UA standard, which point to nodes which do no not have a reference to a ModellingRule Node.
    In order to do this the function also requires tables of the nodes and references which describe
    the types in the references table.

    Instance Declarations require having a HasModellingRule reference to a Modelling Rule.
    If this function is used in a type namespace, it will produce non-hierarchical reference
    types which do not point to InstanceDeclarations.

    Args:
        references (pd.DataFrame): A table of references on which to perform the sub-setting on.
        type_references (pd.DataFrame): A table of references describing the type namespace.
        type_nodes (pd.DataFrame): A table of nodes describing the type namespace.

    Returns:
        pd.DataFrame: Dataframe containing the list of all non-hierarchical references where the
            target is a node which does not have a reference to a Modelling Rule node.

    """

    non_hierarchical_refs = non_hierarchical_references(
        inst_references=references,
        type_references=type_references,
        type_nodes=type_nodes,
    )
    has_modelling_rule = has_modelling_rule_references(
        inst_references=references,
        type_references=type_references,
        type_nodes=type_nodes,
    )[["Src"]].rename(columns={"Src": "id"})
    has_modelling_rule = has_modelling_rule.set_index("id")
    non_hierarchical_refs_has_no_modelling_rule = non_hierarchical_refs.set_index(
        "Trg", drop=False
    )
    trg_no_modelling_rule = ~non_hierarchical_refs_has_no_modelling_rule.index.isin(
        has_modelling_rule.index
    )
    non_hierarchical_refs_has_no_modelling_rule = (
        non_hierarchical_refs_has_no_modelling_rule[trg_no_modelling_rule].copy()
    )
    return non_hierarchical_refs_has_no_modelling_rule[
        ["Src", "Trg", "ReferenceType"]
    ].copy()


def find_relatives(
    nodes: pd.DataFrame,
    nodes_key_col: str,
    edges: pd.DataFrame,
    relative_type: str,
    cutoff: int = None,
    keep_paths: bool = False,
) -> pd.DataFrame:
    """
    Find the relatives to the provided nodes

    The

    Parameters:
    nodes (pd.DataFrame): The dataframe with nodes to find relatives from.
    nodes_key_col (str): The key column to use for indexing in the nodes DataFrame
    edges(pd.DataFrame): The edges to follow in the traverse
    cutoff (int): Max number of steps to go for the traverse
    keep_paths(bool):

    Returns:


    """
    new_nodes = nodes.copy()
    orig_cols = nodes.columns.values.tolist()
    new_nodes[0] = nodes[nodes_key_col]
    new_nodes["len_path"] = 0
    new_nodes = new_nodes.set_index(0, drop=False)
    path_cols = [0]
    if relative_type.lower().startswith("a"):
        col_src = "Trg"
        col_trg = "Src"
    else:
        col_src = "Src"
        col_trg = "Trg"
    # Set the index on reference column based on relativetype(direction)
    edge_join = edges.set_index(col_src)
    result = [new_nodes]
    i = 1
    while (new_nodes.shape[0] > 0) and ((cutoff is None) or (cutoff >= i)):
        joined_nodes = new_nodes.join(edge_join, how="inner")
        #
        new_nodes = joined_nodes[orig_cols + [col_trg] + path_cols].copy()
        if len(new_nodes) > 0:
            if keep_paths:
                new_nodes[i] = new_nodes[col_trg]
                path_cols.append(i)
            else:
                new_nodes[0] = new_nodes[col_trg]
            new_nodes["len_path"] = i
            new_nodes = new_nodes.set_index(col_trg)
            result.append(new_nodes)
        i = i + 1
    if len(result) > 1:
        resconc = pd.concat(result).reset_index()
    else:
        resconc = result[0]

    resconc["end"] = resconc[0]
    if keep_paths:
        for i in range(1, i - 1):
            resconc["end"] = resconc[i].combine_first(resconc["end"])

    return resconc


def resolve_id_from_nodeid(nodes: pd.DataFrame, nodeid: str):
    nodeid_instance = parse_nodeid(nodeid)
    nodes = nodes.set_index("NodeId")
    theid = nodes.loc[nodeid_instance, "id"]
    return theid


def resolve_ids_from_browsenames(nodes: pd.DataFrame, browsenames: List[str]):
    nodes = nodes.set_index("BrowseName")
    hst_id = nodes.loc[browsenames, "id"].reset_index(drop=True)
    return hst_id


def has_subtype_references(
    references: pd.DataFrame, type_nodes: pd.DataFrame, id_col: str
) -> pd.DataFrame:
    hst_id = type_nodes.loc[
        (type_nodes["NodeClass"] == "UAReferenceType")
        & (type_nodes["BrowseName"] == "HasSubtype"),
        id_col,
    ].iloc[0]
    subtype = references[references["ReferenceType"] == hst_id].copy()
    return subtype


def has_type_definition_references(
    references: pd.DataFrame, type_nodes: pd.DataFrame
) -> pd.DataFrame:
    htd_id = type_nodes.loc[
        (type_nodes["NodeClass"] == "UAReferenceType")
        & (type_nodes["BrowseName"] == "HasTypeDefinition"),
        "id",
    ].iloc[0]
    htd = references[references["ReferenceType"] == htd_id].copy()
    return htd


def has_property_references(
    inst_references: pd.DataFrame,
    type_references: pd.DataFrame,
    type_nodes: pd.DataFrame,
) -> pd.DataFrame:
    propref_id = type_nodes.loc[
        (type_nodes["NodeClass"] == "UAReferenceType")
        & (type_nodes["BrowseName"] == "HasProperty"),
        "id",
    ].iloc[0]
    prop_ref = constrain_to_reference_type(
        inst_references, type_nodes, type_references, [propref_id]
    )
    return prop_ref


def has_modelling_rule_references(
    inst_references: pd.DataFrame,
    type_references: pd.DataFrame,
    type_nodes: pd.DataFrame,
) -> pd.DataFrame:
    propref_id = type_nodes.loc[
        (type_nodes["NodeClass"] == "UAReferenceType")
        & (type_nodes["BrowseName"] == "HasModellingRule"),
        "id",
    ].iloc[0]
    prop_ref = constrain_to_reference_type(
        inst_references, type_nodes, type_references, [propref_id]
    )
    return prop_ref


def hierarchical_references(
    inst_references: pd.DataFrame,
    type_references: pd.DataFrame,
    type_nodes: pd.DataFrame,
) -> pd.DataFrame:
    hierref_id = type_nodes.loc[
        (type_nodes["NodeClass"] == "UAReferenceType")
        & (type_nodes["BrowseName"] == "HierarchicalReferences"),
        "id",
    ].iloc[0]
    prop_ref = constrain_to_reference_type(
        inst_references, type_nodes, type_references, [hierref_id]
    )
    return prop_ref


def non_hierarchical_references(
    inst_references: pd.DataFrame,
    type_references: pd.DataFrame,
    type_nodes: pd.DataFrame,
) -> pd.DataFrame:
    """This function expects a table of references on which to find the subset
    of references containing only references of the subtype "NonHierarchicalReferences"
    in OPC UA standard. In order to do this the function also requires tables of the
    nodes and references which describe the types in the references table.

    Args:
        inst_references (pd.DataFrame): A table of references on which to perform the sub-setting on.
        type_references (pd.DataFrame): A table of references describing the type namespace.
        type_nodes (pd.DataFrame): A table of nodes describing the type namespace.

    Returns:
        pd.DataFrame: Dataframe containing the list of all non-hierarchical references.

    """

    non_hierref_id = type_nodes.loc[
        (type_nodes["NodeClass"] == "UAReferenceType")
        & (type_nodes["BrowseName"] == "NonHierarchicalReferences"),
        "id",
    ].iloc[0]
    prop_ref = constrain_to_reference_type(
        inst_references, type_nodes, type_references, [non_hierref_id]
    )
    return prop_ref


def fast_transitive_closure(references: pd.DataFrame) -> pd.DataFrame:
    # Quickly and memory-efficiently compute the transitive closure of a reference
    # Transitive closure: https://en.wikipedia.org/wiki/Transitive_closure
    # We do not consider ReferenceType or store underlying paths in this closure
    # The references are converted to a matrix where each unique Src/Trg is assigned a column and row
    # We have row(Src), col(Trg) == 1 if there is a reference from Src to Trg, 0 otherwise
    # A sparse matrix representation is used in order to reduce memory consumption.
    # The transitive closure is given by iterative dot products of this matrix with itself until we reach a fixed point
    # At the end of each iteration, a cell is converted to 1 if it is above 0 or 0 otherwise.
    # This is in order to ensure that the algorithm terminates
    assert (
        references["Src"] == references["Trg"]
    ).sum() == 0, "There should be no references r such that r(n,n)"
    codes, uniques = pd.factorize(pd.concat([references["Src"], references["Trg"]]))
    src_codes = codes[0 : references.shape[0]]
    trg_codes = codes[references.shape[0] :]
    data = np.ones(references.shape[0])
    sparmat = coo_matrix(
        (data, (src_codes, trg_codes)),
        shape=(uniques.shape[0], uniques.shape[0]),
        dtype=bool,
    )
    sparmat = sparmat + eye(uniques.shape[0])
    fixedp = False
    while not fixedp:
        before = sparmat
        sparmat = sparmat.dot(sparmat) > 0
        if before.sum(axis=None) == sparmat.sum(axis=None):
            fixedp = True

    sparmat = sparmat - eye(uniques.shape[0])
    sparmat = sparmat.tocoo()
    src = sparmat.row
    trg = sparmat.col
    src_nids = uniques.take(src)
    trg_nids = uniques.take(trg)
    hierarchy = pd.DataFrame({"Src": src_nids, "Trg": trg_nids})
    return hierarchy


def fast_hierarchy_transitive_closure(
    inst_references: pd.DataFrame,
    constrained_references: pd.DataFrame,
    type_nodes: pd.DataFrame,
    type_references: pd.DataFrame,
    interesting_types: pd.Series,
) -> pd.DataFrame:
    # Computes a navigation hierarchy between instances of interesting_types
    # The navigation hierarchy is first computed between all types (given references in inst_references)
    # Then, it is restricted to only concern instances of interesting_types
    constrained_references = constrained_references[["Src", "Trg", "ReferenceType"]]
    hierarchy = fast_transitive_closure(constrained_references)

    types = type_nodes[type_nodes["NodeClass"].str.endswith("Type")].set_index(
        "BrowseName"
    )
    interesting_ids_types = types.loc[pd.Index(interesting_types), "id"].to_list()

    sub_inter_types = subtypes_of_nodes(
        interesting_ids_types, type_nodes, type_references
    )
    inter_subtypes_index = pd.Index(sub_inter_types["subtype"])
    htd = has_type_definition_references(inst_references, type_nodes).set_index("Trg")
    htd = htd[htd.index.isin(inter_subtypes_index)]
    inst_inter_type_index = pd.Index(htd["Src"])

    hierarchy = hierarchy.set_index("Src")
    hierarchy = hierarchy[hierarchy.index.isin(inst_inter_type_index)].reset_index()
    hierarchy = hierarchy.set_index("Trg")
    hierarchy = hierarchy[hierarchy.index.isin(inst_inter_type_index)].reset_index()
    return hierarchy


def signal_variables(
    instance_references: pd.DataFrame,
    type_nodes: pd.DataFrame,
    type_references: pd.DataFrame,
) -> pd.DataFrame:
    reftypes = type_nodes[type_nodes["NodeClass"] == "UAVariableType"]
    signalvar_type = reftypes.loc[
        reftypes["BrowseName"] == "BaseDataVariableType", "id"
    ].iloc[0]
    signalvar_types = subtypes_of_nodes([signalvar_type], type_nodes, type_references)
    signalvar_types_index = pd.Index(signalvar_types["subtype"])
    inst_htd = has_type_definition_references(instance_references, type_nodes)
    inst_htd = inst_htd.set_index("Trg")
    signal_vars = inst_htd[inst_htd.index.isin(signalvar_types_index)]["Src"]

    return signal_vars


def property_variables(
    instance_references: pd.DataFrame,
    type_nodes: pd.DataFrame,
    type_references: pd.DataFrame,
) -> pd.DataFrame:
    reftypes = type_nodes[type_nodes["NodeClass"] == "UAVariableType"]
    propvar_type = reftypes.loc[reftypes["BrowseName"] == "PropertyType", "id"].iloc[0]
    propvar_types = subtypes_of_nodes([propvar_type], type_nodes, type_references)
    propvar_types_index = pd.Index(propvar_types["subtype"])
    inst_htd = has_type_definition_references(instance_references, type_nodes)
    inst_htd = inst_htd.set_index("Trg")
    prop_vars = inst_htd[inst_htd.index.isin(propvar_types_index)]["Src"]

    return prop_vars


def typing_transitive_reflexive(
    type_nodes: pd.DataFrame, type_references: pd.DataFrame
) -> pd.DataFrame:
    subtype = has_subtype_references(type_references, type_nodes, "id")[["Src", "Trg"]]
    # Add transitive closure
    subtype_closure = fast_transitive_closure(subtype)
    unique_types = pd.concat([type_references["Src"], type_references["Trg"]]).unique()
    # Add reflexive closure
    subtype_closure = pd.concat(
        [subtype_closure, pd.DataFrame({"Src": unique_types, "Trg": unique_types})]
    )
    return subtype_closure


def subtypes_of_nodes(
    types_to_subtype: Union[pd.Series, List[str]],
    type_nodes: pd.DataFrame,
    type_references: pd.DataFrame,
) -> pd.DataFrame:
    # Computes the closure
    closure = typing_transitive_reflexive(type_nodes, type_references)
    closure = closure.set_index("Src")
    typeindex = pd.Index(types_to_subtype)
    closure = closure[closure.index.isin(typeindex)].reset_index()
    closure = closure.rename(columns={"Trg": "subtype", "Src": "type"})

    return closure


def closest_parent(
    inst_nodes: pd.DataFrame,
    inst_references: pd.DataFrame,
    type_nodes: pd.DataFrame,
    type_references: pd.DataFrame,
    interesting_types: List[str],
    uavar_type: str,
) -> pd.DataFrame:
    sub_inter_types = subtypes_of_nodes(interesting_types, type_nodes, type_references)
    inter_subtypes_index = pd.Index(sub_inter_types["subtype"])
    htd = has_type_definition_references(inst_references, type_nodes).set_index("Trg")
    htd = htd[htd.index.isin(inter_subtypes_index)]
    inst_inter_type_index = pd.Index(htd["Src"])

    # Remove incoming edges to objects of desired type
    inst_references = inst_references.set_index("Trg")
    inst_references = inst_references[
        ~inst_references.index.isin(inst_inter_type_index)
    ].reset_index()

    signals = signal_variables(inst_references, type_nodes, type_references)
    signals_index = pd.Index(signals)

    if uavar_type == "Property":
        inst_references = inst_references.set_index("Trg")
        inst_references = inst_references[
            ~inst_references.index.isin(signals_index)
        ].reset_index()

    inst_references_trans = fast_transitive_closure(inst_references[["Src", "Trg"]])
    inst_nodes_ns_index = inst_nodes[["id", "ns"]].set_index("id")
    inst_references_trans = (
        inst_references_trans.set_index("Src")
        .join(inst_nodes_ns_index)
        .reset_index()
        .rename(columns={"index": "Src", "ns": "Src_ns"})
    )
    inst_references_trans = (
        inst_references_trans.set_index("Trg")
        .join(inst_nodes_ns_index)
        .reset_index()
        .rename(columns={"index": "Trg", "ns": "Trg_ns"})
    )

    if uavar_type == "Property":
        properties = property_variables(inst_references, type_nodes, type_references)
        inst_references_trans = inst_references_trans.set_index("Trg")
        inst_references_trans = inst_references_trans[
            inst_references_trans.index.isin(pd.Index(properties))
        ]
        inst_references_trans = inst_references_trans.reset_index().set_index("Src")
        inst_references_trans = inst_references_trans[
            (
                inst_references_trans.index.isin(signals_index)
                | inst_references_trans.index.isin(inst_inter_type_index)
            )
        ]
        inst_references_trans["is_signal"] = inst_references_trans.index.isin(
            signals_index
        )
        inst_references_trans = inst_references_trans.reset_index()
        inst_references_trans = inst_references_trans.sort_values(
            by=["Trg", "is_signal"], ascending=[True, False]
        )
        inst_references_trans = inst_references_trans.drop_duplicates(keep="first")
        inst_references_trans = inst_references_trans.drop(columns="is_signal")
        return inst_references_trans.rename(
            columns={
                "Src": "instance",
                "Trg": "property",
                "Src_ns": "instance_ns",
                "Trg_ns": "property_ns",
            }
        )

    if uavar_type == "Signal":
        inst_references_trans = inst_references_trans.set_index("Trg")
        inst_references_trans = inst_references_trans[
            inst_references_trans.index.isin(signals_index)
        ]
        inst_references_trans = inst_references_trans.reset_index().set_index("Src")
        inst_references_trans = inst_references_trans[
            inst_references_trans.index.isin(inst_inter_type_index)
        ].reset_index()
        inst_references_trans = inst_references_trans.drop_duplicates(keep="first")
        return inst_references_trans.rename(
            columns={
                "Src": "instance",
                "Trg": "variable",
                "Src_ns": "instance_ns",
                "Trg_ns": "variable_ns",
            }
        )


def constrain_to_reference_type(
    inst_references: pd.DataFrame,
    type_nodes: pd.DataFrame,
    type_references: pd.DataFrame,
    reference_types: List[str],
) -> pd.DataFrame:
    subtypes_of_reference = subtypes_of_nodes(
        reference_types, type_nodes, type_references
    )
    inst_references = inst_references.set_index("ReferenceType")
    inst_references = inst_references[
        inst_references.index.isin(pd.Index(subtypes_of_reference["subtype"]))
    ]
    inst_references = inst_references.reset_index()
    return inst_references



from opcua_tools import ua_models


def test_ensure_that_required_models_are_hashable():
    required_models_set = set()
    required_model = ua_models.UARequiredModel(
        model_uri="http://opcfoundation.org/UA/IEC61850-7-3",
        publication_date="2023-01-01T00:00:00Z",
        version="2.0",
    )

    required_models_set.add(required_model)

    assert len(required_models_set) == 1



import pandas as pd

from opcua_tools import constants, ua_data_types
from opcua_tools.validator import exceptions

UALISTOF_CLASS_NAME = ua_data_types.UAListOf.__name__
VALUE_VALIDATION_ERROR_MESSAGE = (
    "Invalid Value for rows with the following display names: {}."
)


def validate_values_in_df(
    df_to_validate: pd.DataFrame, data_type_nodes: pd.DataFrame
) -> pd.DataFrame:
    """
    Validate that the value is valid for the corresponding DataType for all rows in a DataFrame.
    """
    if df_to_validate.empty or "Value" not in df_to_validate.columns:
        return df_to_validate

    not_empty_values_mask = ~df_to_validate["Value"].isna()
    if "NodeClass" in df_to_validate.columns:
        df_with_rows_to_validate = df_to_validate[
            (df_to_validate["NodeClass"] == "UAVariable") & not_empty_values_mask
        ]
    else:
        df_with_rows_to_validate = df_to_validate[not_empty_values_mask]

    if df_with_rows_to_validate.empty:
        return df_to_validate

    if df_with_rows_to_validate["DataType"].isna().any():
        raise exceptions.ValidationError(
            "UAVariables has no DataType! "
            f"Values: {df_to_validate.loc[df_to_validate['DataType'].isna(), 'Value'].to_list()}"
        )

    df_with_rows_to_validate["StringifiedValue"] = df_with_rows_to_validate[
        "Value"
    ].astype(str)
    df_with_rows_to_validate["StringifiedOpcuaToolsClass"] = df_with_rows_to_validate[
        "StringifiedValue"
    ].str.split("(", n=1, expand=True)[0]
    data_types_mapping = constants.DATA_TYPES_MAPPING
    df_with_rows_to_validate["StringifiedDatatypeClass"] = df_with_rows_to_validate[
        "StringifiedOpcuaToolsClass"
    ].replace(data_types_mapping)
    data_int_type_to_data_type_class_mapping = {
        str(k): v for k, v in data_type_nodes["DisplayName"].to_dict().items()
    }

    df_with_rows_to_validate["ExpectedDataType"] = (
        df_with_rows_to_validate["DataType"]
        .astype("str")
        .replace(data_int_type_to_data_type_class_mapping)
    )

    df_with_rows_to_validate["IsValidValue"] = (
        df_with_rows_to_validate["StringifiedDatatypeClass"]
        == df_with_rows_to_validate["ExpectedDataType"]
    )

    simple_data_types_names = list(data_types_mapping.values())
    skip_validation_mask = (
        df_with_rows_to_validate["StringifiedDatatypeClass"] == UALISTOF_CLASS_NAME
    ) | ~df_with_rows_to_validate["ExpectedDataType"].isin(simple_data_types_names)

    potentially_invalid_df = df_with_rows_to_validate[~skip_validation_mask]

    if not potentially_invalid_df["IsValidValue"].all():
        does_not_contain_ua_enumaration_only_mask = ~(
            potentially_invalid_df["StringifiedDatatypeClass"] == "UAEnumeration"
        )
        if not potentially_invalid_df[does_not_contain_ua_enumaration_only_mask].empty:
            invalid_display_names = df_with_rows_to_validate.loc[
                ~df_with_rows_to_validate["IsValidValue"], "DisplayName"
            ]
            invalid_display_names_as_list = invalid_display_names.values.tolist()
            raise exceptions.ValidationError(
                VALUE_VALIDATION_ERROR_MESSAGE.format(invalid_display_names_as_list)
            )
    return df_to_validate



import os

import pytest
from definitions import get_project_root

from opcua_tools import UAEnumeration, UAGraph, UAInt32
from opcua_tools.nodes_manipulation import transform_ints_to_enums
from opcua_tools.nodeset_parser import parse_xml_files


@pytest.fixture(scope="session")
def ua_graph(paper_example_path):
    path_to_xmls = str(paper_example_path)
    ua_graph = UAGraph.from_path(path_to_xmls)
    transform_ints_to_enums(ua_graph)
    return ua_graph


def test_basic_ua_graph_from_path(ua_graph):
    """From checking the data previously, we found that there are 251 UAVariables
    which point to enumeration data types. Out of these 84 of then do not contain
    an Int64 value, but 'None' instead. This means the number of UAEnumeration
    classes in the Values column, should be 167 after running the
    transform_ints_to_enums()."""

    ua_enumeration_mask = ua_graph.nodes["Value"].apply(
        lambda value: isinstance(value, UAEnumeration)
    )
    ua_enumeration_sum = ua_enumeration_mask.sum()

    assert ua_enumeration_sum == 168


def test_enum_ua_graph_xml_encode(ua_graph):
    """Test checks that the UAEnumeration is correctly encoded as an Int32
    when 'xml_encode()' is run. This should occur as the UAEnumeration is an
    inherited class of UAInt32.
    """
    output_folder = get_project_root() / "tests" / "output"
    output_file_path = str(output_folder / "nodeset_enum_output.xml")

    # Creating output folder if it does not exist
    if not os.path.exists(str(output_folder)):
        os.makedirs(str(output_folder))

    ua_graph.write_nodeset(output_file_path, "http://prediktor.com/paper_example")

    # Reading the file back in, without transforming to enumerations
    parse_dict = parse_xml_files([output_file_path])
    nodes = parse_dict["nodes"]

    # Getting nodes which correspond to the enum test node
    enum_test_node = nodes[nodes["DisplayName"] == "EnumTest"]

    # It should be read correctly read as UAInt32 and thus
    # the output was writing was to UAInt32
    assert type(enum_test_node["Value"].values[0]) == UAInt32



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
from typing import Dict

import pandas as pd
import xmltodict

from opcua_tools.ua_data_types import (
    UAEnumeration,
    UAExtensionObject,
    UAInt32,
    UAListOf,
    UALocalizedText,
)
from opcua_tools.ua_graph import UAGraph

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def create_enum_dict_from_enum_tuples(row: pd.DataFrame) -> Dict[int, str]:
    """Gets the row for the enum node and will attempt to parse the tuple in the
    'Value' column to create and return a dictionary of the enumeration definition.

    Args:
       row (pd.DataFrame): A row of the enumeration which contains a definition with a tuple to parse

    Return:
       A dictionary which defines the enumeration

    """

    # The enum definition can be stored in different ways. The two versions observed
    # is as an unparsed UAExtentionObject in xml or as UALocalizedText.
    enum_tuple = row["Value"]
    enum_dict = dict()
    for index, content in enumerate(enum_tuple):
        if isinstance(content, UAExtensionObject):
            ua_structure = content.body
            xml_string = ua_structure.value
            xml_dict = xmltodict.parse(xml_string)

            # This notation and searching is to ignore xmlns handling
            enum_value_dict = [
                val for key, val in xml_dict.items() if "EnumValueType" in key
            ]
            value = [val for key, val in enum_value_dict[0].items() if "Value" in key][
                0
            ]
            if value.isdigit():
                value = int(value)
            enum_value_dict = [
                val for key, val in xml_dict.items() if "EnumValueType" in key
            ][0]
            display_name_dict = [
                val for key, val in enum_value_dict.items() if "DisplayName" in key
            ][0]
            text = [val for key, val in display_name_dict.items() if "Text" in key][0]
            enum_dict[value] = text
        elif isinstance(content, UALocalizedText):
            value = index
            text = content.text
            enum_dict[value] = text
        else:
            raise ValueError(f"The enum content was not supported: {content}")

    if not enum_dict:
        raise ValueError("The enum dict was not correctly iterated over")

    return enum_dict


def instantiate_enum_class(row: pd.DataFrame) -> UAEnumeration:
    """For a single row from a modified nodes dataframe, which represents an enum value,
    it will produce an UAEnumeration class based on; the Int provided in the
    row, the name of the enumeration which is references by the datatype, and
    the corresponding string value found in the enums definition.

    Args:
       row (pd.DataFrame): The row from nodes with additional 'EnumDict' and 'EnumName' cols.

    Return:
       UAEnumeration class created from the row.

    """

    # In some cases the Value is empty for some enums and should do
    # nothing in this case
    if pd.isna(row["Value"]) is True:
        return pd.NA

    # Extracting the actual numeric value as UAInt32 class
    if isinstance(row["Value"], UAInt32):
        ua_int = row["Value"].value
    elif isinstance(row["Value"], UAListOf):
        ua_int = row["Value"].value[0].value
    else:
        raise ValueError(f"row['Value']: {row['Value']} has not been handled properly")

    # Find the enumeration type name for the variable
    string = row["EnumDict"][ua_int]
    data_type_name = row["EnumName"]

    enum_class = UAEnumeration(value=ua_int, string=string, name=data_type_name)

    return enum_class


def create_enum_definition_table(
    ua_graph: UAGraph, enum_data_types: pd.Series
) -> pd.DataFrame:
    """This function will create an pd.DataFrame containing the id of the DataType which
    the UAVariable for an enumeration class has, the BrowseName of the enumeration class,
    and the actual enumeration definition for that class.

    Args:
       ua_graph (UAGraph): The complete UAGraph with necessary definitions.
       enum_data_types (pd.Series): The column of DataTypes which are Enumeration definitions

    Return:
       pd.DataFrame containing the enumeration definition tables.

    """
    # Subsetting the 'nodes' table to the subset of datatypes
    enum_table = ua_graph.nodes[ua_graph.nodes["id"].isin(enum_data_types)]
    enum_table = enum_table[["id", "BrowseName"]]

    # Getting the references table pointing from the nodes in enum_table
    # via a HasProperty definition to the actual defintion in the node
    enum_references = ua_graph.references[
        ua_graph.references["Src"].isin(enum_data_types)
    ]
    has_property_id = ua_graph.reference_type_by_browsename("HasProperty")
    enum_references = enum_references[
        enum_references["ReferenceType"] == has_property_id
    ]
    enum_references = enum_references.drop(["ReferenceType"], axis=1)

    # Joining the enum_references and enum_table
    enum_table = enum_table.set_index("id", drop=False)
    enum_references = enum_references.set_index("Src", drop=True)
    enum_table = enum_table.join(enum_references, how="left")

    # Joining the enum_table's 'Trg' column to nodes to get the definition
    nodes_subset = ua_graph.nodes[["id", "Value"]].set_index("id")
    enum_table = enum_table.set_index("Trg", drop=True)
    enum_table = enum_table.join(nodes_subset, how="left")

    enum_table["Value"] = enum_table["Value"].apply(lambda row: row.value)
    enum_table["EnumDict"] = enum_table.apply(
        lambda row: create_enum_dict_from_enum_tuples(row), axis=1
    )
    enum_table = enum_table.drop(["Value"], axis=1)

    enum_table = enum_table.rename(columns={"BrowseName": "EnumName"})

    return enum_table


def transform_ints_to_enums(ua_graph: UAGraph):
    """Transforms the integer values in the Value column
    for the Enum data types, to the Enumeration python class. It modifies
    the input UAGraph directly

    Args:
       ua_graph (UAGraph): UAGraph object to transform

    """

    # Extracting nodes for readability
    nodes = ua_graph.nodes.copy()

    # Have to first if the are any Enumeration browsenames in nodes dataframe
    if not (nodes["BrowseName"] == "Enumeration").any():
        return

    # Have to find all children of the Enumeration class
    enumeration_id = ua_graph.data_type_by_browsename("Enumeration")
    enumeration_children = ua_graph.get_neighboring_nodes_by_id(
        enumeration_id, "outgoing"
    )

    # Extracting UAVariables which datatypes are Enumeration subtypes
    enum_data_type_ids = enumeration_children["Trg"]
    enum_nodes = nodes.loc[nodes["DataType"].isin(enum_data_type_ids)]
    enum_nodes = enum_nodes[enum_nodes["NodeClass"] == "UAVariable"]

    logger.info(f"The size of enum_nodes is {enum_nodes.shape}")
    enum_data_types = enum_nodes["DataType"]
    enum_def_table = create_enum_definition_table(ua_graph, enum_data_types)

    # Have to join the enum_definition_table to enum_nodes table
    enum_def_table = enum_def_table.set_index("id", drop=True)
    enum_nodes["DataType"] = enum_nodes["DataType"].astype(str).astype(int)
    enum_nodes = enum_nodes.set_index("DataType", drop=False)
    enum_nodes = enum_nodes.join(enum_def_table, how="left")

    # Apply function to create UAEnumeration in Values column
    enum_nodes["Value"] = enum_nodes.apply(
        lambda row: instantiate_enum_class(row), axis=1
    )
    enum_nodes = enum_nodes.set_index("id", drop=False)
    enum_nodes = enum_nodes.drop(["EnumName", "EnumDict"], axis=1)

    logger.info("Finished working through the enum_nodes")
    # Putting the modified nodes back into the nodes dataframe by index
    nodes.loc[enum_nodes.index, :] = enum_nodes[:]

    ua_graph.nodes = nodes
    logger.info("Finished transforming the integers to enums")



import os.path

from definitions import get_project_root

from opcua_tools import nodeset_parser


def test_get_namespace_data_from_file():
    example_xml_path = os.path.join(
        get_project_root(),
        "tests",
        "testdata",
        "parser",
        "Opc.Ua.IEC61850-7-4.NodeSet2.xml",
    )

    ns_data = nodeset_parser.get_namespace_data_from_file(example_xml_path)

    assert ns_data == {
        "included_namespaces": {
            "http://opcfoundation.org/UA/",
            "http://opcfoundation.org/UA/IEC61850-7-3",
        },
        "name": "http://opcfoundation.org/UA/IEC61850-7-4",
    }


def test_get_namespaca_data_should_have_empty_namespaces_data_regardless_file_name():
    example_xml_path = os.path.join(
        get_project_root(),
        "tests",
        "testdata",
        "parser",
        "Opc.Ua.NodeSet2.Services.xml",
    )

    ns_data = nodeset_parser.get_namespace_data_from_file(example_xml_path)

    assert ns_data == {
        "included_namespaces": set(),
        "name": "http://opcfoundation.org/UA/",
    }



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pandas as pd
import pytest
from definitions import get_project_root

from opcua_tools.ua_data_types import UAEURange
from opcua_tools.ua_graph import UAGraph
from opcua_tools.value_parser import parse_boolean


@pytest.fixture(scope="session")
def boolean_test_data():
    df_values = [
        ["Boolean", "false", False],
        ["Boolean", "False", False],
        ["Boolean", "true", True],
        ["Boolean", "True", True],
    ]
    data = pd.DataFrame(df_values, columns=["tag_type", "input", "solution"])
    return data


@pytest.fixture(scope="session")
def ua_graph(paper_example_path):
    ua_graph = UAGraph.from_path(str(paper_example_path))
    return ua_graph


def test_parse_boolean(boolean_test_data):
    """Ensuring that the parse_boolean function in value_parser.py
    will properly deal with upper and lower case representations of
    strings in python."""

    data = boolean_test_data.copy()
    data["dataclass"] = data["input"].apply(lambda value: parse_boolean(value))
    data["output"] = data["dataclass"].apply(lambda bool_class: bool_class.value)
    pd.testing.assert_series_equal(data["output"], data["solution"], check_names=False)


def test_proper_initiation_of_eurange_class_when_reading_xml(ua_graph):
    """Ensuring that the parse EURange parsing in value_parser.py
    will properly read the EURange in the XML."""

    eurange_nodes = ua_graph.nodes[ua_graph.nodes["DisplayName"] == "EURange"]
    non_none_eurange_nodes = eurange_nodes[~(eurange_nodes["Value"].isnull())]
    non_none_eurange_values = non_none_eurange_nodes["Value"]
    assert non_none_eurange_values.apply(
        lambda value: isinstance(value, UAEURange)
    ).all()


def test_ua_graph_xml_encode_for_eurange(ua_graph):
    """Test checks that the UAEURange is correctly encoded when
    'xml_encode()' is run."""
    output_folder = get_project_root() / "tests" / "output"
    output_file_path = str(output_folder / "nodeset_eurange_output.xml")

    # Creating output folder if it does not exist
    if not os.path.exists(str(output_folder)):
        os.makedirs(str(output_folder))

    ua_graph.write_nodeset(
        output_file_path,
        "http://prediktor.com/paper_example",
    )

    path_to_tests = get_project_root() / "tests"
    expected_file_path = str(
        path_to_tests / "expected" / "value_parser" / "expected_nodeset_eurange.xml"
    )

    expected_eurange_xml_lines = []
    with open(expected_file_path) as expected_file:
        for line in expected_file:
            if "ns=1;i=40" in line or "ns=1;i=31" in line:
                expected_eurange_xml_lines.append(line)

    output_eurange_xml_lines = []
    with open(output_file_path) as output_file:
        for line in output_file:
            if "ns=1;i=40" in line or "ns=1;i=31" in line:
                output_eurange_xml_lines.append(line)

    assert set(output_eurange_xml_lines) == set(expected_eurange_xml_lines)



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from definitions import get_project_root

import opcua_tools as ot
from opcua_tools.nodeset_generator import (
    denormalize_nodes_nodeids,
    denormalize_references_nodeids,
)


def test_values_correct(paper_example_path):
    parse_dict = ot.parse_xml_dir(str(paper_example_path))

    nodes = parse_dict["nodes"]
    references = parse_dict["references"]
    lookup_df = parse_dict["lookup_df"]
    nodes = denormalize_nodes_nodeids(nodes, lookup_df)
    denormalize_references_nodeids(references, lookup_df)
    nodes = nodes.drop(columns=["id"])
    nodes = nodes.sort_values(by=nodes.columns.values.tolist()).reset_index(drop=True)

    nodes = nodes[nodes["ns"] == 1]  # Select only example namespace.

    output_path = (
        get_project_root() / "tests" / "expected" / "paper_example" / "nodes.csv"
    )
    nodes.to_csv(str(output_path), index=False)



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import math
import re
from abc import ABC, abstractmethod
from base64 import b64encode
from dataclasses import astuple, dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Any, ByteString, Optional, Tuple, Union
from xml.sax.saxutils import escape

import numpy as np
import pandas as pd


class NodeIdType(Enum):
    NUMERIC = "i"
    STRING = "s"
    GUID = "g"
    OPAQUE = "b"

    def __repr__(self):
        return "NodeIdType" + "." + self.name


class VariantType(Enum):
    """The possible types of a variant."""

    Null = 0
    Boolean = 1
    SByte = 2
    Byte = 3
    Int16 = 4
    UInt16 = 5
    Int32 = 6
    UInt32 = 7
    Int64 = 8
    UInt64 = 9
    Float = 10
    Double = 11
    String = 12
    DateTime = 13
    Guid = 14
    ByteString = 15
    XmlElement = 16
    NodeId = 17
    ExpandedNodeId = 18
    StatusCode = 19
    QualifiedName = 20
    LocalizedText = 21
    ExtensionObject = 22
    DataValue = 23
    Variant = 24
    DiagnosticInfo = 25


UAXMLNS_ATTRIB = 'xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd"'


@dataclass(eq=True, frozen=True)
class UAData(ABC):
    pass

    @abstractmethod
    def xml_encode(self, include_xmlns: bool) -> str:
        pass

    def __lt__(self, other):
        return lt(self, other)

    def __gt__(self, other):
        return gt(self, other)

    def __le__(self, other):
        return le(self, other)

    def __ge__(self, other):
        return ge(self, other)


@dataclass(eq=True, frozen=True)
class DataTypeField(UAData):
    name: str
    value: str
    description: str

    def xml_encode(self, include_xmlns: bool) -> str:
        x = '<Field Name="' + self.name + '" Value="' + self.value + '"'
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">"
        x += "<Description>" + self.description + "</Description>"
        x += "</Field>"
        return x


@dataclass(eq=True, frozen=True)
class DataTypeDefinition(UAData):
    name: str
    fields: Tuple[DataTypeField, ...]

    def xml_encode(self, include_xmlns: bool) -> str:
        encodedvalues = [v.xml_encode(include_xmlns=False) for v in self.fields]
        x = '<Definition Name="' + self.name + '"'
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + "\n".join(encodedvalues) + "</Definition>"
        return x


@dataclass(eq=True, frozen=True)
class UABuiltIn(UAData):
    def resolve_builtin_type_number(self) -> int:
        class_name = self.__class__.__name__
        if class_name.startswith("UA"):
            class_name = class_name[2:]
        if class_name in VariantType.__members__:
            return VariantType[class_name].value
        else:
            raise ValueError(f"Unknown built-in type: {class_name}")


@dataclass(eq=True, frozen=True)
class UAInteger(UABuiltIn):
    value: int = pd.NA

    def __post_init__(self):
        if not pd.isna(self.value) and not isinstance(self.value, int):
            raise TypeError("Integer value must be an int")


@dataclass(eq=True, frozen=True)
class UASignedInteger(UAInteger):
    value: int = pd.NA


@dataclass(eq=True, frozen=True)
class UAUnsignedInteger(UAInteger):
    value: int = pd.NA

    def __post_init__(self):
        if not pd.isna(self.value) and not isinstance(self.value, int):
            raise TypeError("Integer value must be an int")
        if not pd.isna(self.value) and self.value < 0:
            raise ValueError("UnsignedInteger value cannot be negative")


@dataclass(eq=True, frozen=True)
class UASByte(UASignedInteger):
    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<SByte"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + (str(self.value) if not pd.isna(self.value) else "") + "</SByte>"
        return x

    def json_encode(self) -> Union[str, None]:
        if pd.isna(self.value):
            return None
        else:
            return str(self.value)


@dataclass(eq=True, frozen=True)
class UAByte(UAUnsignedInteger):
    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<UAByte"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + (str(self.value) if not pd.isna(self.value) else "") + "</UAByte>"
        return x

    def json_encode(self) -> Union[str, None]:
        if pd.isna(self.value):
            return None
        else:
            return str(self.value)


@dataclass(eq=True, frozen=True)
class UAInt16(UASignedInteger):
    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<Int16"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + (str(self.value) if not pd.isna(self.value) else "") + "</Int16>"
        return x

    def json_encode(self) -> Union[str, None]:
        if pd.isna(self.value):
            return None
        else:
            return str(self.value)


@dataclass(eq=True, frozen=True)
class UAUInt16(UAUnsignedInteger):
    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<UInt16"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + (str(self.value) if not pd.isna(self.value) else "") + "</UInt16>"
        return x

    def json_encode(self) -> Union[str, None]:
        if pd.isna(self.value):
            return None
        else:
            return str(self.value)


@dataclass(eq=True, frozen=True)
class UAInt32(UASignedInteger):
    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<Int32"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + (str(self.value) if not pd.isna(self.value) else "") + "</Int32>"
        return x

    def json_encode(self) -> Union[str, None]:
        if pd.isna(self.value):
            return None
        else:
            return str(self.value)


@dataclass(eq=True, frozen=True)
class UAUInt32(UAUnsignedInteger):
    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<UInt32"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + (str(self.value) if not pd.isna(self.value) else "") + "</UInt32>"
        return x

    def json_encode(self) -> Union[str, None]:
        if pd.isna(self.value):
            return None
        else:
            return str(self.value)


@dataclass(eq=True, frozen=True)
class UAInt64(UASignedInteger):
    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<Int64"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + (str(self.value) if not pd.isna(self.value) else "") + "</Int64>"
        return x

    def json_encode(self) -> Union[str, None]:
        # Int64 and UInt64 are to be formatted as number encoded
        # as a JSON string according to spec
        if pd.isna(self.value):
            return None
        else:
            return '"' + str(float(self.value)) + '"'


@dataclass(eq=True, frozen=True)
class UAUInt64(UAUnsignedInteger):
    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<UInt64"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + (str(self.value) if not pd.isna(self.value) else "") + "</UInt64>"
        return x

    def json_encode(self) -> Union[str, None]:
        # Int64 and UInt64 are to be formatted as number encoded
        # as a JSON string according to spec
        if pd.isna(self.value):
            return None
        else:
            return '"' + str(float(self.value)) + '"'


@dataclass(eq=True, frozen=True)
class UAFloatingPoint(UABuiltIn):
    value: float = pd.NA

    def __post_init__(self):
        if self.value is not None:
            if not pd.isna(self.value) and not isinstance(
                self.value, (float, int, Decimal)
            ):
                raise TypeError(
                    "Floating Point Numbers value must be either be a float, int, or Decimal"
                )
        if isinstance(self.value, pd._libs.missing.NAType) or self.value is None:
            object.__setattr__(self, "value", pd.NA)
        else:
            object.__setattr__(self, "value", float(self.value))

    def json_encode(self) -> [str, None]:
        # According to the spec, special values are to be encoded as JSON
        # strings in the following manner
        special_floats_or_decimals = {
            None: None,
            pd.NA: None,
            float("inf"): '"Infinity"',
            float("-inf"): '"-Infinity"',
            float("nan"): "NaN",
        }
        if self.value in special_floats_or_decimals:
            return special_floats_or_decimals[self.value]
        elif math.isnan(self.value):
            return '"NaN"'
        else:
            return str(self.value)


@dataclass(eq=True, frozen=True)
class UAFloat(UAFloatingPoint):
    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<Float"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + (str(self.value) if not pd.isna(self.value) else "") + "</Float>"
        return x


@dataclass(eq=True, frozen=True)
class UADouble(UAFloatingPoint):
    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<Double"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + (str(self.value) if not pd.isna(self.value) else "") + "</Double>"
        return x


@dataclass(eq=True, frozen=True)
class UAString(UABuiltIn):
    value: str = pd.NA

    def __post_init__(self):
        if not pd.isna(self.value) and not isinstance(self.value, str):
            raise TypeError("String value must be a string")
        if not (value := str(self.value)) or pd.isna(self.value):
            object.__setattr__(self, "value", pd.NA)
        else:
            object.__setattr__(self, "value", value)

    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<String"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += (
            ">"
            + (escape(str(self.value)) if not pd.isna(self.value) else "")
            + "</String>"
        )
        return x

    def json_encode(self) -> Union[str, None]:
        if pd.isna(self.value):
            return None
        else:
            return json.dumps(self.value, ensure_ascii=False)


@dataclass(eq=True, frozen=True)
class UADateTime(UABuiltIn):
    value: datetime = pd.NA

    def __post_init__(self):
        if not isinstance(self.value, datetime):
            raise TypeError("DateTime value must be a datetime object")

    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<DateTime"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += (
            ">"
            + (
                self.value.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
                if not pd.isna(self.value)
                else ""
            )
            + "</DateTime>"
        )
        return x

    def json_encode(self) -> Union[str, None]:
        """
        The time format used here is the recommended for RDF graphs
        is the format W3C is recommended. The format is:
        YYYY-MM-DDThh:mm:ss.sssTZD ex. 2023-05-08T14:30:00.000Z
        Is the same recommended format used by the OPC UA JSON encoding.
        """
        return '"' + self.value.strftime("%Y-%m-%dT%H:%M:%S.%fZ") + '"'


@dataclass(eq=True, frozen=True)
class UAGuid(UAString):
    pass


@dataclass(eq=True, frozen=True)
class UAByteString(UABuiltIn):
    value: ByteString = pd.NA

    def __post_init__(self):
        if not pd.isna(self.value) and not isinstance(self.value, ByteString):
            raise TypeError("ByteString value must be a ByteString object")
        object.__setattr__(self, "value", self.value if self.value else pd.NA)

    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<ByteString"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += (
            ">"
            + (b64encode(self.value).decode("utf-8") if not pd.isna(self.value) else "")
            + "</ByteString>"
        )
        return x

    def json_encode(self):
        """ByteStryings are encoded as base64 strings in JSON, are enclosed
        in double quotes, and cannot contain illegal JSON characters.
        """
        if pd.isna(self.value) and not isinstance(self.value, ByteString):
            return None
        else:
            base64_string = b64encode(self.value).decode("utf-8")
            return json.dumps(base64_string, ensure_ascii=False)


@dataclass(eq=True, frozen=True)
class UABoolean(UABuiltIn):
    value: bool = pd.NA

    def __post_init__(self):
        if not pd.isna(self.value) and not isinstance(self.value, bool):
            raise TypeError("Boolean value must be a bool")

    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<Boolean"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += (
            ">"
            + (
                ("true" if self.value is True else "false")
                if not pd.isna(self.value)
                else ""
            )
            + "</Boolean>"
        )
        return x

    def json_encode(self) -> Union[str, None]:
        if pd.isna(self.value):
            return None
        else:
            return str(self.value).lower()


@dataclass(eq=True, frozen=True)
class UAXMLElement(UABuiltIn):
    value: str

    def __post_init__(self):
        if not isinstance(self.value, str):
            raise TypeError("XMLElement value must be a string")

    def xml_encode(self, include_xmlns: bool) -> str:
        return self.value

    def json_encode(self) -> str:
        return json.dumps(self.value, ensure_ascii=False)


@dataclass(eq=True, frozen=True)
class UANodeId(UABuiltIn):
    namespace: int
    nodeid_type: NodeIdType
    value: Union[str, int]

    def __post_init__(self):
        if not isinstance(self.namespace, int):
            raise TypeError("Namespace must be an integer")
        if not isinstance(self.nodeid_type, (NodeIdType, int, str)):
            raise TypeError("NodeIdType must be a NodeIdType, int, or valid string")
        if not isinstance(self.value, (str, int)):
            raise TypeError("Value must be a string")

        # Casting into to Enum if int is passed
        if isinstance(self.nodeid_type, int):
            nodeid_type_symbol = self.nodeid_type_int_to_symbol(self.nodeid_type)
            object.__setattr__(self, "nodeid_type", NodeIdType(nodeid_type_symbol))

        # Casting into to Enum if string is passed
        if isinstance(self.nodeid_type, str):
            object.__setattr__(self, "nodeid_type", NodeIdType(self.nodeid_type))

        # Adding validation that the value is a valid NodeIdType based on the
        # value
        if self.nodeid_type == NodeIdType.NUMERIC:
            # Check if the value is a valid unsigned integer
            if not isinstance(self.value, (int, str)):
                raise TypeError(
                    f"Value must be an unsigned integer or string for NodeIdType {self.nodeid_type.value}"
                )
            if isinstance(self.value, str):
                if not self.value.isdigit():
                    raise TypeError(
                        f"Value must be an unsigned integer for NodeIdType {self.nodeid_type.value}"
                    )
                if self.value.startswith("-"):
                    raise TypeError(
                        f"Value must not start with 0 for NodeIdType {self.nodeid_type.value}"
                    )
                if self.value.startswith("0") and len(self.value) > 1:
                    raise TypeError(
                        f"Value must not start with 0 for NodeIdType {self.nodeid_type.value}"
                    )

    def __str__(self):
        if self.namespace == 0:
            return f"{self.nodeid_type.value}={self.value}"
        else:
            return f"ns={self.namespace};{self.nodeid_type.value}={self.value}"

    @staticmethod
    def nodeid_type_int_to_symbol(nodeid_type_int: int) -> str:
        nodeid_type_int_to_string = {0: "i", 1: "s", 2: "g", 3: "b"}
        if nodeid_type_int in nodeid_type_int_to_string:
            return nodeid_type_int_to_string[nodeid_type_int]
        raise TypeError(f"NodeIdType {nodeid_type_int} is not supported")

    def nodeid_type_value_to_int(self):
        nodeid_type_symbol_to_int = {"i": 0, "s": 1, "g": 2, "b": 3}
        if self.nodeid_type.value in nodeid_type_symbol_to_int:
            return nodeid_type_symbol_to_int[self.nodeid_type.value]
        raise TypeError(f"NodeIdType {self.nodeid_type.value} is not supported")

    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<Identifier"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + self.__str__() + "</Identifier>"
        return x

    def json_encode(self) -> str:
        nodeid_type_int = self.nodeid_type_value_to_int()
        json_content = ""
        # OPC UA Specification indicates Namespace field is omitted if it is 0
        if not self.namespace == 0:
            json_content = '"Namespace":' + str(self.namespace) + ","
        # OPC UA Specification indicates IdentifierType field is omitted if it is 0 (UInt32)
        if not nodeid_type_int == 0:
            json_content += '"IdType":' + str(nodeid_type_int) + ","
        # Properly encoding the identifier value based on datatype
        if self.nodeid_type is NodeIdType.NUMERIC:
            json_content += '"Id":' + str(self.value)
        else:  # "s", "g", "b"
            # If string, Guid identifier or byte string (Opaque) should treat value as string
            json_content += '"Id":' + '"' + str(self.value) + '"'

        return "{" + json_content + "}"


@dataclass(eq=True, frozen=True)
class UAExpandedNodeId(UAString):
    value: str = pd.NA


@dataclass(eq=True, frozen=True)
class UAStatusCode(UABuiltIn):
    value: int


@dataclass(eq=True, frozen=True)
class UADiagnosticInfo(UABuiltIn):
    pass


@dataclass(eq=True, frozen=True)
class UAQualifiedName(UABuiltIn):
    namespace_index: np.uint16
    name: str

    def __post_init__(self):
        if not isinstance(self.namespace_index, int):
            raise TypeError("Namespace index must be an integer")
        if isinstance(self.namespace_index, int):
            if self.namespace_index < 0:
                raise ValueError("Namespace index must be a positive integer")
            object.__setattr__(self, "namespace_index", np.uint16(self.namespace_index))
        if not isinstance(self.name, str):
            raise TypeError("Name must be a string")

    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<QualifiedName"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">"
        x += "<NamespaceIndex>" + str(self.namespace_index) + "</NamespaceIndex>"
        x += "<Name>" + self.name + "</Name></QualifiedName>"
        return x

    def json_encode(self) -> str:
        json = ""
        json += '"Name":"' + self.name + '"'
        if not self.namespace_index == 0:
            json += ',"Uri":' + str(self.namespace_index)
        return "{" + json + "}"


@dataclass(eq=True, frozen=True)
class UALocalizedText(UABuiltIn):
    text: Optional[str] = pd.NA
    locale: Optional[str] = pd.NA

    def __post_init__(self):
        if (not isinstance(self.text, str)) and (not pd.isna(self.text)):
            raise TypeError("text must be a string or pd.NA")
        if (not isinstance(self.locale, str)) and (not pd.isna(self.locale)):
            raise TypeError("locale must be a string or pd.NA")
        if not pd.isna(self.locale):
            if not self.is_valid_locale(self.locale):
                raise ValueError(
                    "locale must be a valid locale string following the IETF RFC 5646 standard"
                )

    @staticmethod
    def is_valid_locale(locale_str: str):
        """Validates the format of the locale string using the IETF RFC 5646 standard
        The following regex is based on the regex provided by the IETF RFC 5646 standard
        and allows for the following formats:
        ^[a-zA-Z]{2,3}: 2-3 letter language code
        (?:[_-][a-zA-Z]{2,3}: optional 2-3 letter script code
        (?:[_-](?:\w{2,8}|\d{3}))?)?: optional 2-8 letter region code or 3 digit country code
        (?:\.[a-zA-Z0-9]{2,8})?$: optional variant code and optional extension code

        Args:
            locale_str (str): locale string to be validated

        Returns:
            bool: True if locale string is valid, False otherwise
        """

        if re.match(
            r"^[a-zA-Z]{2,3}(?:[_-][a-zA-Z]{2,3}(?:[_-](?:\w{2,8}|\d{3}))?)?(?:\.[a-zA-Z0-9]{2,8})?$",
            locale_str,
        ):
            return True
        else:
            return False

    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<LocalizedText"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">"
        x += (
            "<Locale>" + (self.locale if not pd.isna(self.locale) else "") + "</Locale>"
        )
        x += (
            "<Text>" + (escape(self.text) if not pd.isna(self.text) else "") + "</Text>"
        )
        x += "</LocalizedText>"
        return x

    def json_encode(self, input_locale: Optional[str] = None) -> Union[str, None]:
        json_content = ""

        if pd.isna(self.text):
            json_content += '"Text":' + json.dumps("", ensure_ascii=False)
        else:
            json_content += '"Text":' + json.dumps(self.text, ensure_ascii=False)

        if not pd.isna(input_locale):
            json_content += ',"Locale":' + json.dumps(input_locale, ensure_ascii=False)
        elif not pd.isna(self.locale):
            json_content += ',"Locale":' + json.dumps(self.locale, ensure_ascii=False)

        json_content = "{" + json_content + "}"
        return json_content


@dataclass(eq=True, frozen=True)
class UAVariant(UABuiltIn):
    value: Any = pd.NA
    type: VariantType = VariantType.Null

    def __post_init__(self):
        ua_class_type_to_variant_type = {
            UABoolean: VariantType.Boolean,
            UASByte: VariantType.SByte,
            UAByte: VariantType.Byte,
            UAInt16: VariantType.Int16,
            UAUInt16: VariantType.UInt16,
            UAInt32: VariantType.Int32,
            UAEnumeration: VariantType.Int32,
            UAUInt32: VariantType.UInt32,
            UAInt64: VariantType.Int64,
            UAUInt64: VariantType.UInt64,
            UAFloat: VariantType.Float,
            UADouble: VariantType.Double,
            UAString: VariantType.String,
            UADateTime: VariantType.DateTime,
            UAGuid: VariantType.Guid,
            UAByteString: VariantType.ByteString,
            UAXMLElement: VariantType.XmlElement,
            UANodeId: VariantType.NodeId,
            UAExpandedNodeId: VariantType.ExpandedNodeId,
            UAStatusCode: VariantType.StatusCode,
            UAQualifiedName: VariantType.QualifiedName,
            UALocalizedText: VariantType.LocalizedText,
            UAExtensionObject: VariantType.ExtensionObject,
            UADataValue: VariantType.DataValue,
            UAVariant: VariantType.Variant,
            UADiagnosticInfo: VariantType.DiagnosticInfo,
        }
        # Checking that the value is of the correct type
        if not pd.isna(self.value):
            if not isinstance(self.value, tuple(ua_class_type_to_variant_type.keys())):
                raise TypeError(
                    "Value must be one of the valid Built-In types in OPC UA"
                )
        # Checking that the type is of the correct type
        if not isinstance(self.type, VariantType):
            raise TypeError("Type must be of type VariantType")

        if self.type is VariantType.Null and not pd.isna(self.value):
            if ua_class_type_to_variant_type.get(type(self.value)) is not None:
                object.__setattr__(
                    self, "type", ua_class_type_to_variant_type.get(type(self.value))
                )
            else:
                raise ValueError(
                    "Variant type was not provided and could not be inferred from the value type"
                )

    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<Variant"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">"
        if not pd.isna(self.value):
            x += "<Value>"
            if isinstance(self.value, UABuiltIn) and hasattr(self.value, "xml_encode"):
                x += self.value.xml_encode(include_xmlns=include_xmlns)
            else:
                x += escape(str(self.value))
            x += "</Value>"
        x += "</Variant>"
        return x

    def json_encode(self, **kwargs) -> str:
        if pd.isna(self.value):
            return "null"
        if self.type is VariantType.Null:
            return "null"

        if isinstance(self.value, UABuiltIn) and hasattr(self.value, "json_encode"):
            body_json = self.value.json_encode(**kwargs)
        elif isinstance(self.value, str):
            body_json = '"' + str(self.value) + '"'
        elif isinstance(self.value, (int, float)):
            body_json = str(self.value)
        elif isinstance(self.value, bool):
            body_json = '"' + str(self.value).lower() + '"'
        else:
            raise TypeError(f"Variant value type {type(self.value)} is not supported")

        if body_json is None:
            return "null"

        json_content = "{" + '"Type":' + str(self.type.value) + ","
        json_content += '"Body":' + body_json + "}"
        return json_content


@dataclass(eq=True, frozen=True)
class UADataValue(UABuiltIn):
    pass


@dataclass(eq=True, frozen=True)
class UADecimal(UAData):
    pass


@dataclass(eq=True, frozen=True)
class UAEnumeration(UAInt32):
    string: str = ""
    name: str = ""

    def __post_init__(self):
        if not isinstance(self.string, str):
            raise TypeError("String must be a string")
        if not isinstance(self.name, str):
            raise TypeError("Name must be a string")


@dataclass(eq=True, frozen=True)
class UAArray(UAData):
    pass


@dataclass(eq=True, frozen=True)
class UAStructure(UAData):
    value: Any = pd.NA

    def xml_encode(self, include_xmlns: bool) -> str:
        if pd.isna(self.value):
            return ""
        if not pd.isna(self.value) and hasattr(self.value, "xml_encode"):
            return self.value.xml_encode(include_xmlns)
        else:
            return ""

    def json_encode(self, **kwargs) -> str:
        if pd.isna(self.value):
            return "null"
        if not pd.isna(self.value) and hasattr(self.value, "json_encode"):
            return self.value.json_encode(**kwargs)
        else:
            return "null"


@dataclass(eq=True, frozen=True)
class UAStructureOptionalField(UAData):
    pass


@dataclass(eq=True, frozen=True)
class UAExtensionObject(UABuiltIn):
    type_nodeid: UANodeId = pd.NA
    body: Union[UAByteString, UAStructure, UAXMLElement] = pd.NA
    encoding_json: int = pd.NA

    def __post_init__(self):
        if not isinstance(self.type_nodeid, UANodeId):
            raise TypeError("ExtensionObject must have a type_nodeid")
        if not isinstance(self.body, (UAByteString, UAStructure, UAXMLElement)):
            raise TypeError(
                "ExtensionObject must have a body of type UAByteString, UAStructure or UAXMLElement"
            )

        instance_map = {
            UAStructure: 0,
            UAByteString: 1,
            UAXMLElement: 2,
            0: pd.NA,
        }
        object.__setattr__(
            self, "encoding_json", instance_map.get(self.body.__class__, pd.NA)
        )

    def xml_encode(self, include_xmlns: bool) -> str:
        x = "<ExtensionObject"
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">"
        x += (
            "<TypeId>"
            + (
                self.type_nodeid.xml_encode(include_xmlns=False)
                if not pd.isna(self.type_nodeid)
                else ""
            )
            + "</TypeId>"
        )
        x += (
            "<Body>"
            + (
                self.body.xml_encode(include_xmlns=False)
                if not pd.isna(self.body)
                else ""
            )
            + "</Body>"
        )
        x += "</ExtensionObject>"
        return x

    def json_encode(self, **kwargs) -> str:
        if pd.isna(self.body) or not hasattr(self.body, "json_encode"):
            return "null"
        if self.body.json_encode(**kwargs) is None:
            return "null"

        json = ""
        json += '"TypeId":' + self.type_nodeid.json_encode() + ","

        # The body can be a ByteString, Structure or an XML Element and
        # has to have a json_encode method.
        if self.body.json_encode(**kwargs) is None:
            return "null"
        elif self.body.json_encode(**kwargs) is not None:
            json += '"Body":' + self.body.json_encode(**kwargs)
        else:
            raise TypeError(
                f"ExtensionObject body type {type(self.body)} is not supported"
            )

        # OPC UA Spec: If the Body is None, NULL, or 0 (Structure), the Encoding field shall be omitted.
        if (not pd.isna(self.encoding_json)) and (self.encoding_json != 0):
            json += ","
            json += '"Encoding":' + str(self.encoding_json)
        return "{" + json + "}"


@dataclass(eq=True, frozen=True)
class UAEUInformation(UAData):
    display_name: UALocalizedText = pd.NA
    description: UALocalizedText = pd.NA
    unit_id: int = pd.NA
    namespace_uri: str = pd.NA

    def __post_init__(self):
        if not isinstance(self.display_name, UALocalizedText) and isinstance(
            self.display_name, str
        ):
            object.__setattr__(
                self, "display_name", UALocalizedText(text=self.display_name)
            )
        elif isinstance(self.display_name, UALocalizedText) or pd.isna(
            self.display_name
        ):
            object.__setattr__(self, "display_name", self.display_name)
        else:
            raise TypeError("display_name must be of type UALocalizedText")

        if not isinstance(self.description, UALocalizedText) and isinstance(
            self.description, str
        ):
            object.__setattr__(
                self, "description", UALocalizedText(text=self.description)
            )
        elif isinstance(self.description, UALocalizedText) or pd.isna(self.description):
            object.__setattr__(self, "description", self.description)
        else:
            raise TypeError("description must be of type UALocalizedText")

        if not isinstance(self.unit_id, int):
            raise TypeError("unit_id must be of type int")
        object.__setattr__(self, "unit_id", self.unit_id)

        if not isinstance(self.namespace_uri, str):
            raise TypeError("namespace_uri must be of type str")
        object.__setattr__(self, "namespace_uri", self.namespace_uri)

    def xml_encode(self, include_xmlns: bool) -> str:
        body_contents = "<EUInformation"
        if include_xmlns:
            body_contents += ' xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd"'
        body_contents += ">"
        body_contents += "<NamespaceUri>" + self.namespace_uri + "</NamespaceUri>"
        body_contents += "<UnitId>" + str(self.unit_id) + "</UnitId>"
        body_contents += "<DisplayName>"
        body_contents += (
            "<Locale>"
            + (
                self.display_name.locale
                if not pd.isna(self.display_name.locale)
                else "en"
            )
            + "</Locale>"
        )
        body_contents += (
            "<Text>"
            + (self.display_name.text if not pd.isna(self.display_name.text) else "")
            + "</Text>"
        )
        body_contents += "</DisplayName>"
        body_contents += "<Description>"
        body_contents += (
            "<Locale>"
            + (
                self.description.locale
                if not pd.isna(self.description.locale)
                else "en"
            )
            + "</Locale>"
        )
        body_contents += (
            "<Text>"
            + (self.description.text if not pd.isna(self.description.text) else "")
            + "</Text>"
        )
        body_contents += "</Description>"
        body_contents += "</EUInformation>"

        return body_contents

    def json_encode(self, **kwargs) -> Union[str, None]:
        if (
            self.display_name.json_encode(input_locale=kwargs.get("input_locale"))
            is None
            or self.description.json_encode(input_locale=kwargs.get("input_locale"))
            is None
        ):
            return None

        json_content = ""
        decoded_display_name = self.display_name.json_encode(
            input_locale=kwargs.get("input_locale", None)
        )
        json_content += (
            '{"DisplayName":' + decoded_display_name + ","
            if decoded_display_name
            else '""' + ","
        )
        decoded_description = self.description.json_encode(
            input_locale=kwargs.get("input_locale", None)
        )
        json_content += (
            '"Description":' + decoded_description + ","
            if decoded_description
            else '""' + ","
        )
        json_content += '"UnitId":' + str(self.unit_id) + ","
        json_content += '"NamespaceUri":' + json.dumps(
            self.namespace_uri, ensure_ascii=False
        )
        json_content += "}"
        return json_content


@dataclass(eq=True, frozen=True)
class UAEngineeringUnits(UAData):
    ua_eu_information: UAEUInformation

    def __init__(
        self,
        display_name: UALocalizedText,
        description: UALocalizedText,
        unit_id: int,
        namespace_uri: str,
    ):
        object.__setattr__(
            self,
            "ua_eu_information",
            UAEUInformation(
                display_name=display_name,
                description=description,
                unit_id=unit_id,
                namespace_uri=namespace_uri,
            ),
        )

    def __repr__(self):
        """
        Fixed object representation, so it can be reinstantiated with performing
        eval() on the object's representation string.
        """
        ua_eu_information_param_name = list(self.__dataclass_fields__.keys())[0]
        ua_eu_information_repr = repr(getattr(self, ua_eu_information_param_name))
        ua_eu_information = getattr(self, ua_eu_information_param_name)
        return ua_eu_information_repr.replace(
            ua_eu_information.__class__.__name__, self.__class__.__name__, 1
        )

    def xml_encode(self, include_xmlns: bool) -> str:
        ua_type_nodeid = UANodeId(0, NodeIdType.NUMERIC, "888")
        ua_extension_object = UAExtensionObject(
            type_nodeid=ua_type_nodeid,
            body=UAStructure(value=self.ua_eu_information),
        )
        return ua_extension_object.xml_encode(include_xmlns)

    def json_encode(self, **kwargs) -> str:
        ua_type_nodeid = UANodeId(0, NodeIdType.NUMERIC, "888")
        ua_extension_object = UAExtensionObject(
            type_nodeid=ua_type_nodeid,
            body=UAStructure(value=self.ua_eu_information),
        )
        return ua_extension_object.json_encode(**kwargs)


@dataclass(eq=True, frozen=True)
class UARange(UAData):
    low: float
    high: float

    def __post_init__(self):
        if self.low > self.high:
            raise ValueError("low must be less than or equal to high")
        if isinstance(self.low, int):
            object.__setattr__(self, "low", float(self.low))
        if isinstance(self.high, int):
            object.__setattr__(self, "high", float(self.high))

    def xml_encode(self, include_xmlns: bool) -> str:
        body_contents = "<Range"
        if include_xmlns:
            body_contents += ' xmlns="http://opcfoundation.org/UA/2008/02/Types.xsd"'
        body_contents += ">"
        body_contents += f"<Low>{self.low}</Low>"
        body_contents += f"<High>{self.high}</High>"
        body_contents += "</Range>"
        return body_contents

    def json_encode(self) -> str:
        json = ""
        json += '"Low":' + UADouble(self.low).json_encode() + ","
        json += '"High":' + UADouble(self.high).json_encode()
        return "{" + json + "}"


@dataclass(eq=True, frozen=True)
class UAEURange(UAData):
    ua_range: UARange

    def __init__(self, low: float, high: float):
        if not isinstance(low, (float, int)):
            raise TypeError("low must be of type float")
        if not isinstance(high, (float, int)):
            raise TypeError("high must be of type float")

        object.__setattr__(self, "ua_range", UARange(low=float(low), high=float(high)))

    def __repr__(self):
        """
        Fixed object representation, so it can be reinstantiated with performing
        eval() on the object's representation string.
        """
        ua_range_param_name = list(self.__dataclass_fields__.keys())[0]
        ua_range_repr = repr(getattr(self, ua_range_param_name))
        ua_range = getattr(self, ua_range_param_name)
        return ua_range_repr.replace(
            ua_range.__class__.__name__, self.__class__.__name__, 1
        )

    def xml_encode(self, include_xmlns: bool) -> str:
        ua_type_nodeid = UANodeId(0, NodeIdType.NUMERIC, "885")
        ua_extension_object = UAExtensionObject(
            type_nodeid=ua_type_nodeid, body=UAStructure(value=self.ua_range)
        )

        return ua_extension_object.xml_encode(include_xmlns)

    def json_encode(self) -> str:
        ua_type_nodeid = UANodeId(0, NodeIdType.NUMERIC, "885")
        ua_extension_object = UAExtensionObject(
            type_nodeid=ua_type_nodeid, body=UAStructure(value=self.ua_range)
        )
        return ua_extension_object.json_encode()


@dataclass(eq=True, frozen=True)
class UAUnion(UAData):
    pass


@dataclass(eq=True, frozen=True)
class UAMessage(UAData):
    pass


@dataclass(eq=True, frozen=True)
class UAEmpty(UAData):
    pass


@dataclass(eq=True, frozen=True)
class UAListOf(UAData):
    value: Tuple[UAData, ...]
    typename: str

    def __post_init__(self):
        first_element_type = type(self.value[0])
        if any(
            not isinstance((_element := element), first_element_type)
            for element in self.value
        ):
            raise TypeError(
                f"UAListOf must contain only one type of UAData. {first_element_type} != {type(_element)}"
            )

    def xml_encode(self, include_xmlns: bool) -> str:
        encodedvalues = []
        if not pd.isna(self.value):
            encodedvalues = [v.xml_encode(include_xmlns=False) for v in self.value]

        x = "<ListOf" + self.typename + " "
        if include_xmlns:
            x += " " + UAXMLNS_ATTRIB
        x += ">" + "".join(encodedvalues) + "</ListOf" + self.typename + ">"
        return x

    def json_encode(self) -> str:
        """Extracts the values of a UAListOf object and hard codes the values within it
        in order to be able to create the proper json_encoding as it aggregates all of the
        values in a list.

        Returns:
            str: A string representing the json encoding of the UAListOf object.
        """
        try:
            str_list = [str(element.value) for element in self.value]
        except AttributeError:
            print("UAListOf json_encode error")
        ua_list_string = "[" + ",".join(str_list) + "]"

        return (
            f"""{{"Type":{VariantType[self.typename].value},"Body":{ua_list_string}}}"""
        )


def ge(u1: UAData, u2: UAData):
    return le(u1=u2, u2=u1)


def le(u1: UAData, u2: UAData):
    if u2 is None:
        return True
    ty1 = type(u1)
    ty2 = type(u2)
    if ty1 != ty2:
        return ty1.__name__ <= ty2.__name__

    t1 = str(astuple(u1))
    t2 = str(astuple(u2))

    return t1 <= t2


def gt(u1: UAData, u2: UAData):
    return lt(u1=u2, u2=u1)


def lt(u1: UAData, u2: UAData):
    if u2 is None:
        return True
    ty1 = type(u1)
    ty2 = type(u2)
    if ty1 != ty2:
        return ty1.__name__ < ty2.__name__

    t1 = str(astuple(u1))
    t2 = str(astuple(u2))

    return t1 < t2



# Copyright 2021 Prediktor AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import gc
import logging
import os
import time
from datetime import datetime
from io import StringIO
from typing import List, Optional, Union
from xml.sax.saxutils import escape

import lxml.etree as ET
import numpy as np
import pandas as pd
import pytz

from opcua_tools import memory_optimizer, ua_models
from opcua_tools.ua_data_types import UANodeId
from opcua_tools.validator import value_validator

PATH_HERE = os.path.dirname(__file__)


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
memory_optimizer.replace_default_pandas_del_method()


simplevariants = {
    "Boolean",
    "SByte",
    "Byte",
    "Int16",
    "UInt16",
    "Int32",
    "UInt32",
    "Int64",
    "UInt64",
    "Float",
    "Double",
    "String",
    "DateTime",
    "Guid",
    "ByteString",
}


def create_header_xml(
    namespaces,
    serialize_namespace,
    models: List[ua_models.UAModel],
    xmlns_dict: Optional[dict] = None,
    last_modified: Optional[datetime] = None,
    publication_date: Optional[datetime] = None,
):

    if not xmlns_dict:
        xmlns_dict = {
            "xsd": "http://www.w3.org/2001/XMLSchema",
            "xsi": "http://www.w3.org/2001/XMLSchema-instance",
            None: "http://opcfoundation.org/UA/2011/03/UANodeSet.xsd",
        }

    if not last_modified:
        last_modified = datetime.now(tz=pytz.UTC)

    if not publication_date:
        publication_date = datetime.now(tz=pytz.UTC)

    prefixes = ""
    for key in xmlns_dict:
        prefix = "xmlns"
        if key:
            prefix += ":{}".format(key)
        prefixes += ' {}="{}"'.format(prefix, xmlns_dict[key])

    namespaceheader = ""
    if len(namespaces) > 1:
        namespaceheader += "<NamespaceUris>\n"
        for i, n in enumerate(namespaces):
            if i < 1:
                continue

            namespaceheader += "<Uri>{}</Uri>\n".format(n)
        namespaceheader += "</NamespaceUris>\n"

    model_uri = namespaces[serialize_namespace]
    try:
        model = next(model for model in models if model.model_uri == model_uri)
    except StopIteration:
        model = ua_models.UAModel(
            model_uri=model_uri,
            publication_date=publication_date.isoformat(),
            version=None,
        )

    required_models = create_required_models(model)
    default_version = "1.0.0"
    version = model.version if model.version is not None else default_version

    return """<?xml version="1.0" encoding="utf-8"?>
<UANodeSet LastModified="{}" {}>
{}
<Models>
    <Model ModelUri="{}" PublicationDate="{}" Version="{}">{}</Model>
</Models>
<Aliases></Aliases>
""".format(
        last_modified.isoformat(),
        prefixes,
        namespaceheader,
        model_uri,
        publication_date.isoformat(),
        version,
        required_models,
    )


def generate_references_xml(nodes, references):
    """
    :param nodes:
    :param references:
    :return:
    """

    # references = references.reset_index()
    nodes["target_same_ns"] = True
    references = references.set_index("Trg").join(
        nodes.set_index("NodeId")[["target_same_ns"]], how="left"
    )
    references = references.reset_index().rename(columns={"index": "Trg"})
    nodes.drop(columns="target_same_ns", inplace=True)
    references["target_same_ns"] = ~references["target_same_ns"].isna()

    # Default code references on target node
    references["NodeId"] = references["Trg"]

    # References where target is not in same namespace are encoded on source
    references.loc[~references["target_same_ns"], "NodeId"] = references.loc[
        ~references["target_same_ns"], "Src"
    ]

    # Outgoing reference when target is not in same nodeset
    references.loc[~references["target_same_ns"], "xml"] = (
        '<Reference ReferenceType="'
        + references.loc[~references["target_same_ns"], "ReferenceType"]
        + '">'
        + references.loc[~references["target_same_ns"], "Trg"]
        + "</Reference>"
    )

    # Incoming reference only when target is in same nodeset to be generated
    references.loc[references["target_same_ns"], "xml"] = (
        '<Reference ReferenceType="'
        + references.loc[references["target_same_ns"], "ReferenceType"]
        + '"  IsForward="false">'
        + references.loc[references["target_same_ns"], "Src"]
        + "</Reference>"
    )

    references = references.drop(columns=["target_same_ns"])
    return references


def encode_values(nodes):
    is_variable = nodes["NodeClass"] == "UAVariable"
    if "Value" in nodes.columns.values:
        has_value = ~nodes["Value"].isna()
        should_encode = is_variable & has_value
        nodes.loc[should_encode, "EncodedValue"] = nodes.loc[
            should_encode, "Value"
        ].map(lambda x: x.xml_encode(include_xmlns=True))
    else:
        nodes["EncodedValue"] = np.nan


def generate_nodes_xml(
    nodes: pd.DataFrame,
    references: pd.DataFrame,
    lookup_df: pd.DataFrame,
    data_type_nodes: pd.DataFrame,
):
    assert (
        nodes["BrowseNameNamespace"].isna().sum() == 0
    ), "Should not have missing BrowseNameNamespaces"

    replacer = lambda x: x.map(escape)
    nodes["DisplayName"] = replacer(nodes["DisplayName"])
    nodes["BrowseName"] = replacer(nodes["BrowseName"])
    nodes["Description"] = replacer(nodes["Description"])
    nodes["NodeId"] = replacer(nodes["NodeId"].map(str))

    nodes = value_validator.validate_values_in_df(nodes, data_type_nodes)
    nodes = denormalize_nodes_nodeids(nodes, lookup_df=lookup_df)
    references = denormalize_references_nodeids(references, lookup_df=lookup_df)

    references["Src"] = replacer(references["Src"].map(str))
    references["Trg"] = replacer(references["Trg"].map(str))
    references["ReferenceType"] = replacer(references["ReferenceType"].map(str))

    nodes = nodes.copy()

    encode_values(nodes)
    encode_definitions(nodes)
    nodes["nodexml"] = "<" + nodes["NodeClass"] + ' NodeId="' + nodes["NodeId"] + '"'
    if ["SymbolicName"] in nodes.columns.values:
        hasSymbolicName = ~nodes["SymbolicName"].isna()
        nodes.loc[hasSymbolicName, "nodexml"] = (
            nodes.loc[hasSymbolicName, "nodexml"]
            + ' SymbolicName="'
            + nodes.loc[hasSymbolicName, "SymbolicName"]
            + '" '
        )
    nodes["nodexml"] = (
        nodes["nodexml"]
        + ' BrowseName="'
        + nodes["BrowseNameNamespace"].astype(str)
        + ":"
        + nodes["BrowseName"]
        + '" '
    )
    for a in [
        "DataType",
        "ValueRank",
        "AccessLevel",
        "UserAccessLevel",
        "IsAbstract",
        "Symmetric",
        "ParentNodeId",
        "ArrayDimensions",
        "MinimumSamplingInterval",
        "MethodDeclarationId",
        "EventNotifier",
        "Historizing",
        "WriteMask",
    ]:
        if a in nodes.columns.values:
            notna = ~nodes[a].isna()
            haslen = nodes[a].astype(str).str.len() > 0
            if a in ("IsAbstract", "Symmetric", "Historizing"):
                use_value = nodes.loc[notna & haslen, a].astype(str).str.lower()
            else:
                use_value = nodes.loc[notna & haslen, a].astype(str)

            nodes.loc[notna & haslen, "nodexml"] = (
                nodes.loc[notna & haslen, "nodexml"] + a + '="' + use_value + '" '
            )

    nodes["nodexml"] = nodes["nodexml"] + ">"
    nodes["nodexml"] = (
        nodes["nodexml"] + "<DisplayName>" + nodes["DisplayName"] + "</DisplayName>"
    )
    has_description = ~nodes["Description"].isna()
    nodes.loc[has_description, "nodexml"] = (
        nodes.loc[has_description, "nodexml"]
        + "<Description>"
        + nodes.loc[has_description, "Description"]
        + "</Description>"
    )

    new_references = generate_references_xml(nodes, references).reset_index()
    nodes_tojoin = nodes.set_index("NodeId").join(
        new_references.set_index("NodeId")[["xml"]]
    )

    nodes_tojoin["xml"] = nodes_tojoin["xml"].fillna("")
    nodes_tojoin = (
        nodes_tojoin.reset_index()
        .groupby("NodeId")
        .agg({"xml": lambda x: "".join(x.values)})[["xml"]]
    )
    nodes = nodes.set_index("NodeId").join(nodes_tojoin).reset_index()
    nodes["nodexml"] = (
        nodes["nodexml"] + "<References>" + nodes["xml"].fillna("") + "</References>"
    )
    has_encoded_value = ~nodes["EncodedValue"].isna()
    nodes.loc[has_encoded_value, "nodexml"] = (
        nodes.loc[has_encoded_value, "nodexml"]
        + "<Value>"
        + nodes.loc[has_encoded_value, "EncodedValue"]
        + "</Value>"
    )
    if "Definition" in nodes.columns.values:
        has_encoded_definition = ~nodes["EncodedDefinition"].isna()
        nodes.loc[has_encoded_definition, "nodexml"] = (
            nodes.loc[has_encoded_definition, "nodexml"]
            + nodes.loc[has_encoded_definition, "EncodedDefinition"]
        )

    nodes["nodexml"] = nodes["nodexml"] + "</" + nodes["NodeClass"] + ">"

    # Delete pandas DataFrames manually to force memory release
    del replacer
    del references
    del new_references
    del nodes_tojoin
    del lookup_df
    gc.collect()

    return nodes["nodexml"].astype(str)


def encode_definitions(nodes: pd.DataFrame):
    if "Definition" in nodes.columns.values:
        has_definition = ~nodes["Definition"].isna()
        nodes.loc[has_definition, "EncodedDefinition"] = nodes.loc[
            has_definition, "Definition"
        ].map(lambda x: x.xml_encode(include_xmlns=False))


def create_nodeset2_file(
    nodes: pd.DataFrame,
    references: pd.DataFrame,
    models: List[ua_models.UAModel],
    namespaces: List[str],
    serialize_namespace: int,
    filename_or_stringio: Union[str, StringIO] = "nodeset2.xml",
    xmlns_dict=None,
    last_modified: Optional[datetime] = None,
    publication_date: Optional[datetime] = None,
):
    nodes["ns"] = nodes["NodeId"].map(lambda x: x.namespace)
    namespaces_in_use = find_namespaces_in_use(
        nodes=nodes, references=references, namespace_index=serialize_namespace
    )
    namespaces_in_use.sort()
    data_type_nodes = nodes[nodes["NodeClass"] == "UADataType"].copy()
    nodes = nodes[nodes["ns"].map(lambda x: x in (namespaces_in_use))].copy()
    original_namespaces = namespaces
    namespaces = [namespaces[i] for i in namespaces_in_use]
    reindex_nodeids_browsenames(nodes, original_namespaces, namespaces)
    nodes["ns"] = nodes["NodeId"].map(lambda x: x.namespace)
    lookup_df = create_lookup_df(nodes)

    header = create_header_xml(
        namespaces,
        serialize_namespace,
        models,
        xmlns_dict=xmlns_dict,
        last_modified=last_modified,
        publication_date=publication_date,
    )
    start_time = time.time()
    logger.info("Creating nodeset2xml-node-string")

    nodes = nodes[nodes["ns"] == serialize_namespace].copy()
    # References restrict themselves
    nodes_df = generate_nodes_xml(nodes, references, lookup_df, data_type_nodes)
    end_time = time.time()
    logger.info(f"Creating nodeset2xml-string took {str(end_time - start_time)}")
    logger.info("Writing nodeset2xml")
    start_time = time.time()

    outstr = header + "\n".join(nodes_df.values) + "\n" + "</UANodeSet>"

    if type(filename_or_stringio) == str:
        with open(filename_or_stringio, "w", encoding="utf-8") as f:
            f.write(outstr)
    else:
        filename_or_stringio.write(outstr)
    end_time = time.time()

    # Delete pandas DataFrames manually to force memory release
    del nodes_df
    del nodes
    del references
    del lookup_df
    gc.collect()
    logger.info(f"Writing nodeset2xml-file took: {str(end_time - start_time)}")


def reindex_nodeids_browsenames(
    nodes: pd.DataFrame, original_namespaces: List[str], namespaces: List[str]
):
    namespace_map = {original_namespaces.index(n): i for i, n in enumerate(namespaces)}
    nodes["NodeId"] = nodes["NodeId"].map(
        lambda x: copy_nodeid_with_new_namespace(x, namespace_map[x.namespace])
    )
    nodes["BrowseNameNamespace"] = (
        nodes["BrowseNameNamespace"].map(namespace_map).astype(pd.Int32Dtype())
    )


def copy_nodeid_with_new_namespace(uanodeid: UANodeId, new_ns: int):
    return UANodeId(
        namespace=new_ns, value=uanodeid.value, nodeid_type=uanodeid.nodeid_type
    )


def copy_browsename_with_new_namespace(uanodeid: UANodeId, new_ns: int):
    return UANodeId(
        namespace=new_ns, value=uanodeid.value, nodeid_type=uanodeid.nodeid_type
    )


def validate_nodeset2_file(filename: str):
    start_time = time.time()
    tree = ET.parse(PATH_HERE + "/static/UANodeSet.xsd")
    schema = ET.XMLSchema(tree)
    parser = ET.XMLParser(schema=schema)
    try:
        ET.parse(filename, parser)
        logger.info("XML validated by nodeset2 xsd")
    except Exception as e:
        logger.info("XML is invalid according to nodeset2 xsd")
        logger.info("Error occured:")
        logger.info(e)
    end_time = time.time()
    logger.info(f"Validating nodeset2xml-file took: {str(end_time - start_time)}")


def create_lookup_df(nodes):
    return (
        nodes[["id", "NodeId"]]
        .rename(columns={"NodeId": "uniques"})
        .set_index("id", drop=True)
    )


def denormalize_nodes_nodeids(nodes, lookup_df):
    nodecols = ["ParentNodeId", "DataType", "MethodDeclarationId"]

    for c in nodecols:
        if c in nodes.columns.values:
            uniques = lookup_df.rename(columns={"uniques": c}, errors="raise")
            nodes = nodes.set_index(c).join(uniques).reset_index(drop=True)
            nodes[c] = nodes[c].fillna(pd.NA)
    return nodes


def denormalize_references_nodeids(references, lookup_df):
    refcols = ["Src", "Trg", "ReferenceType"]
    for c in refcols:
        uniques = lookup_df.rename(columns={"uniques": c}, errors="raise")
        references = references.set_index(c).join(uniques).reset_index(drop=True)
    return references


def find_namespaces_in_use(nodes, references, namespace_index):
    """Find the namespaces which are in use and the BrowseNameNamespaces
    which are used but already in the list"""
    cs = ["DataType", "ParentNodeId", "MethodDeclarationId"]
    other_nodes = []
    for c in cs:
        if c in nodes.columns.values:
            other_nodes.append(c)
    ids_in_ns = nodes.loc[nodes["ns"] == namespace_index, ["id"] + other_nodes].copy()
    ids_in_ns = ids_in_ns.set_index("id", drop=False)

    references_source_in_ns = references.set_index("Src")
    references_source_in_ns = references_source_in_ns[
        references_source_in_ns.index.isin(ids_in_ns.index)
    ]

    all_ids_set = set(ids_in_ns["id"])

    all_ids_set = all_ids_set.union(set(references_source_in_ns["Trg"]))
    all_ids_set = all_ids_set.union(set(references_source_in_ns["ReferenceType"]))

    references_target_in_ns = references.set_index("Trg")
    references_target_in_ns = references_target_in_ns[
        references_target_in_ns.index.isin(ids_in_ns.index)
    ]

    all_ids_set = all_ids_set.union(set(references_target_in_ns["Src"]))
    all_ids_set = all_ids_set.union(set(references_target_in_ns["ReferenceType"]))

    for c in other_nodes:
        if c in nodes.columns.values:
            all_ids_set = all_ids_set.union(ids_in_ns[c].dropna())

    output_nodes = pd.DataFrame({"id": list(all_ids_set)}).set_index("id")
    nodes = nodes.set_index("id")
    namespaces_in_use = list(
        nodes.loc[nodes.index.isin(output_nodes.index), "ns"].dropna().unique()
    )
    # Adding the BrowseNameNamespaces which are not already in the namespace list
    browsename_namespaces = list(
        nodes.loc[nodes["ns"] == namespace_index, "BrowseNameNamespace"].unique()
    )
    namespaces_in_use = list(set(namespaces_in_use + browsename_namespaces))

    del ids_in_ns
    gc.collect()
    return namespaces_in_use


def create_required_models(model: Optional[ua_models.UAModel]) -> str:
    required_models_str = ""
    if model is None:
        return required_models_str

    for required_model in model.required_models:
        required_model_to_add = """\n        <RequiredModel ModelUri="{}" Version="{}" PublicationDate="{}" />""".format(
            required_model.model_uri,
            required_model.version,
            (
                required_model.publication_date
                if required_model.publication_date is not None
                else datetime.now(tz=pytz.UTC).isoformat()
            ),
        )
        required_models_str = "{}{}".format(
            required_models_str,
            required_model_to_add,
        )
    if model.required_models:
        required_models_str = f"{required_models_str}\n    "

    return required_models_str
